import os
import sys
import re
import json
from urllib.request import Request, urlopen
from urllib.parse import urlparse, parse_qs, unquote, urlencode
from html import unescape
from html.parser import HTMLParser
import websockets
from base64 import b64decode
import string
from random import choice, randint

from pychess import VERSION
from pychess.Utils.const import (
    FEN_START,
    FEN_START_960,
    CRAZYHOUSECHESS,
    FISCHERRANDOMCHESS,
    reprResult,
)
from pychess.Utils.lutils.LBoard import LBoard
from pychess.Utils.lutils.lmove import parseAny, toSAN
from pychess.widgets import newGameDialog
from pychess.System.Log import log

# import pdb
# def _(p):
#     return p
# VERSION = '1.0'
# import ssl
# ssl._create_default_https_context = ssl._create_unverified_context  # Chess24, ICCF


TYPE_GAME, TYPE_STUDY, TYPE_PUZZLE, TYPE_EVENT, TYPE_FEN = range(5)
CHESS960 = "Fischerandom"

CAT_DL = _("Download link")
CAT_HTML = _("HTML parsing")
CAT_API = _("Application programming interface")
CAT_MISC = _("Various techniques")
CAT_WS = _("Websockets")


# Abstract class to download a game from the Internet
class InternetGameInterface:
    # Internal
    def __init__(self):
        """Initialize the common data that can be used in ALL the sub-classes."""
        self.reset()
        self.allow_extra = False
        self.userAgent = "PyChess %s" % VERSION
        self.use_an = True  # To rebuild a readable PGN where possible
        self.allow_octet_stream = False

    def reset(self):
        """Clear the internal variables used to fetch the games."""
        self.id = None
        self.url_type = None
        self.data = None

    def is_enabled(self):
        """Override this method in the sub-class to disable a chess provider temporarily."""
        _, proto = self.get_identity()
        return (proto != CAT_WS) or (
            sys.version_info.major >= 3 and sys.version_info.minor >= 5
        )  # Async WS needs 3.5 for SSL

    def is_async(self):
        """Tell if the chess provider is using asynchronous code."""
        _, proto = self.get_identity()
        return proto == CAT_WS

    def get_description(self):
        """Return the description of the chess provider."""
        name, _ = self.get_identity()
        return "%s" % name

    def get_game_id(self):
        """Return the unique identifier of the game that was detected after a successful call to assign_game().
        The value is None if no game was found earlier."""
        return self.id

    def reacts_to(self, url, host):
        """Return True if the URL belongs to the HOST (possibly equal to *). The sub-domains other than "www" are not supported.
        The method is used to accept any URL when a unique identifier cannot be extracted by assign_game().
        """
        # Verify the hostname
        if url is None:
            return False
        if host != "*":
            parsed = urlparse(url)
            if parsed.netloc.lower() not in ["www." + host.lower(), host.lower()]:
                return False

        # Any page is valid
        self.id = url
        return True

    def json_loads(self, data):
        """Load a JSON and handle the errors.
        The value None is returned when the data are not relevant or misbuilt."""
        try:
            if data in [None, ""]:
                return None
            return json.loads(data)
        except ValueError:
            return None

    def json_field(self, data, path, default=""):
        """Conveniently read a field from a JSON data. The PATH is a key like "node1/node2/key".
        A blank string is returned in case of error."""
        if data in [None, ""]:
            return ""
        keys = path.split("/")
        value = data
        for key in keys:
            if key.startswith("[") and key.endswith("]"):
                try:
                    value = value[int(key[1:-1])]
                except (ValueError, TypeError, IndexError):
                    return ""
            else:
                if key in value:
                    value = value[key]
                else:
                    return ""
        return default if value in [None, ""] else value

    def read_data(self, response):
        """Read the data from an HTTP request and execute the charset conversion.
        The value None is returned in case of error."""
        # Check
        if response is None:
            return None
        bytes = response.read()

        # Decode
        cs = response.info().get_content_charset()
        if cs is not None:
            data = bytes.decode(cs)
        else:
            try:
                data = bytes.decode("utf-8")
            except Exception:
                try:
                    data = bytes.decode("latin-1")
                except Exception:
                    log.debug("Error in the decoding of the data")
                    return None

        # Result
        data = data.replace("\ufeff", "").replace("\r", "").strip()
        if data == "":
            return None
        else:
            return data

    def expand_links(self, links, url):
        """Convert relative paths into full paths."""
        base = urlparse(url)
        for i, link in enumerate(links):
            e = urlparse(link)
            if e.netloc == "":
                if e.path.startswith("/"):
                    link = f"{base.scheme}://{base.netloc}{e.path}"
                else:
                    link = "{}://{}{}/{}".format(
                        base.scheme,
                        base.netloc,
                        "/".join(base.path.split("/")[:-1]),
                        e.path,
                    )
            links[i] = link
        return list(dict.fromkeys(links))  # Without duplicate entries

    def download(self, url, userAgent=False):
        """Download the URL from the Internet.
        The USERAGENT is requested by some websites to make sure that you are not a bot.
        The value None is returned in case of error."""
        # Check
        if url in [None, ""]:
            return None

        # Download
        try:
            log.debug("Downloading game: %s" % url)
            if userAgent:
                req = Request(url, headers={"User-Agent": self.userAgent})
                response = urlopen(req)
            else:
                response = urlopen(url)
            return self.read_data(response)
        except Exception as exception:
            log.debug("Exception raised: %s" % str(exception))
            return None

    def download_list(self, links, userAgent=False):
        """Download and concatenate the URL given in the array LINKS.
        The USERAGENT is requested by some websites to make sure that you are not a bot.
        The number of downloads is limited to 10.
        The downloads that failed are dropped silently.
        The value None is returned in case of no data or error."""
        pgn = ""
        for i, link in enumerate(links):
            data = self.download(link, userAgent)
            if data not in [None, ""]:
                pgn += "%s\n\n" % data
            if i >= 10:  # Anti-flood
                break
        if pgn == "":
            return None
        else:
            return pgn

    def send_xhr(self, url, postData, userAgent=False):
        """Call a target URL by submitting the POSTDATA.
        The USERAGENT is requested by some websites to make sure that you are not a bot.
        The value None is returned in case of error."""
        # Check
        if url in [None, ""] or postData in [None, ""]:
            return None

        # Call data
        try:
            log.debug("Calling API: %s" % url)
            if userAgent:
                req = Request(
                    url,
                    urlencode(postData).encode(),
                    headers={"User-Agent": self.userAgent},
                )
            else:
                req = Request(url, urlencode(postData).encode())
            response = urlopen(req)
            return self.read_data(response)
        except Exception as exception:
            log.debug("Exception raised: %s" % str(exception))
            return None

    def rebuild_pgn(self, game):
        """Return an object in PGN format.
        The keys starting with "_" are dropped silently.
        The key "_url" becomes the first comment.
        The key "_moves" contains the moves.
        The key "_reason" becomes the last comment."""
        # Check
        if game is None or game == "" or "_moves" not in game or game["_moves"] == "":
            return None

        # Convert Chess960 to classical chess depending on the start position
        if "FEN" in game:
            if "Variant" in game:
                if game["Variant"] == CHESS960 and game["FEN"] == FEN_START_960:
                    del game["Variant"], game["SetUp"], game["FEN"]
            else:
                if game["FEN"] == FEN_START:
                    del game["SetUp"], game["FEN"]

        # Header
        pgn = ""
        for e in game:
            if e[:1] != "_" and game[e] not in [None, ""]:
                pgn += f'[{e} "{game[e]}"]\n'
        if pgn == "":
            pgn = '[Annotator "PyChess %s"]\n' % VERSION
        pgn += "\n"

        # Body
        if "_url" in game:
            pgn += "{%s}\n" % game["_url"]
        if "_moves" in game:
            pgn += "%s " % game["_moves"]
        if "_reason" in game:
            pgn += "{%s} " % game["_reason"]
        if "Result" in game:
            pgn += "%s " % game["Result"]
        return pgn.strip()

    def sanitize(self, pgn):
        """Modify the PGN output to comply with the expected format"""
        # Check
        if pgn in [None, ""]:
            return None

        # Verify that it starts with the correct magic character (ex.: "<" denotes an HTML content, "[" a chess game, etc...)
        pgn = pgn.replace("\r", "").strip()
        if not pgn.startswith("["):
            return None

        # Reorganize the spaces to bypass Scoutfish's limitation
        while True:
            lc = len(pgn)
            pgn = pgn.replace("\n\n\n", "\n\n")
            if len(pgn) == lc:
                break

        # Extract the first game
        pos = pgn.find(
            "\n\n["
        )  # TODO Support in-memory database to load several games at once
        if pos != -1:
            pgn = pgn[:pos]

        # Variants
        pgn = pgn.replace('[Variant "Chess"]\n', "")

        # Return the PGN with the local crlf
        return pgn.replace("\n", os.linesep)

    def strip_html(self, input):
        """Remove any HTML mark from the input parameter."""
        rxp = re.compile(r"<\/?[^>]+>", re.IGNORECASE)
        return rxp.sub("", input)

    def is_fen(self, fen):
        """Test if the argument is a FEN position."""
        try:
            rxp = re.compile(
                r"^[kqbnrp1-8\/]+\s[w|b]\s[kq-]+\s[a-h-][1-8]?(\s[0-9]+)?(\s[0-9]+)?$",
                re.IGNORECASE,
            )
            return rxp.match(fen) is not None
        except TypeError:
            return False

    # External
    def get_identity(self):
        """(Abstract) Name and technique of the chess provider."""
        pass

    def assign_game(self, url):
        """(Abstract) Detect the unique identifier of URL."""
        pass

    def download_game(self):
        """(Abstract) Download the game identified earlier by assign_game()."""
        pass


# Lichess.org
class InternetGameLichess(InternetGameInterface):
    def reset(self):
        InternetGameInterface.reset(self)
        self.url_tld = "org"

    def get_identity(self):
        return "Lichess.org", CAT_DL

    def assign_game(self, url):
        # Retrieve the ID of the broadcast
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?lichess\.(org|dev)\/broadcast\/[a-z0-9\-]+\/([a-z0-9]+)[\/\?\#]?",
            re.IGNORECASE,
        )
        m = rxp.match(url)
        if m is not None:
            gid = m.group(3)
            if len(gid) == 8:
                self.url_type = TYPE_STUDY
                self.id = gid
                self.url_tld = m.group(2)
                return True

        # Retrieve the ID of the practice
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?lichess\.(org|dev)\/practice\/[\w\-\/]+\/([a-z0-9]+\/[a-z0-9]+)(\.pgn)?\/?([\S\/]+)?$",
            re.IGNORECASE,
        )
        m = rxp.match(url)
        if m is not None:
            gid = m.group(3)
            if len(gid) == 17:
                self.url_type = TYPE_STUDY
                self.id = gid
                self.url_tld = m.group(2)
                return True

        # Retrieve the ID of the study
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?lichess\.(org|dev)\/study\/([a-z0-9]+(\/[a-z0-9]+)?)(\.pgn)?\/?([\S\/]+)?$",
            re.IGNORECASE,
        )
        m = rxp.match(url)
        if m is not None:
            gid = m.group(3)
            if len(gid) in [8, 17]:
                self.url_type = TYPE_STUDY
                self.id = gid
                self.url_tld = m.group(2)
                return True

        # Retrieve the ID of the puzzle
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?lichess\.(org|dev)\/training\/([0-9]+|daily)[\/\?\#]?",
            re.IGNORECASE,
        )
        m = rxp.match(url)
        if m is not None:
            gid = m.group(3)
            if (gid.isdigit() and gid != "0") or gid == "daily":
                self.url_type = TYPE_PUZZLE
                self.id = gid
                self.url_tld = m.group(2)
                return True

        # Retrieve the ID of the game
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?lichess\.(org|dev)\/(game\/export\/|embed\/)?([a-z0-9]+)\/?([\S\/]+)?$",
            re.IGNORECASE,
        )  # More permissive
        m = rxp.match(url)
        if m is not None:
            gid = m.group(4)
            if len(gid) == 8:
                self.url_type = TYPE_GAME
                self.id = gid
                self.url_tld = m.group(2)
                return True

        # Nothing found
        return False

    def query_api(self, path):
        response = urlopen(
            Request(
                f"https://lichess.{self.url_tld}{path}",
                headers={
                    "X-Requested-With": "XMLHttpRequest",
                    "Accept": "application/vnd.lichess.v4+json",
                },
            )
        )
        bourne = self.read_data(response)
        return self.json_loads(bourne)

    def download_game(self):
        # Check
        if None in [self.id, self.url_tld]:
            return None

        # Logic for the games
        if self.url_type == TYPE_GAME:
            # Download the finished game
            api = self.query_api("/import/master/%s/white" % self.id)
            game = self.json_field(api, "game")
            if "winner" in game:
                url = "https://lichess.{}/game/export/{}?literate=1".format(
                    self.url_tld, self.id
                )
                return self.download(url)
            else:
                if not self.allow_extra and game["rated"]:
                    return None

            # Rebuild the PGN file
            game = {}
            game["_url"] = "https://lichess.{}{}".format(
                self.url_tld, self.json_field(api, "url/round")
            )
            game["Variant"] = self.json_field(api, "game/variant/name")
            game["FEN"] = self.json_field(api, "game/initialFen")
            game["SetUp"] = "1"
            game["White"] = self.json_field(
                api,
                "player/name",
                self.json_field(api, "player/user/username", "Anonymous"),
            )
            game["WhiteElo"] = self.json_field(api, "player/rating")
            game["Black"] = self.json_field(
                api,
                "opponent/name",
                self.json_field(api, "opponent/user/username", "Anonymous"),
            )
            game["BlackElo"] = self.json_field(api, "opponent/rating")
            if self.json_field(api, "clock") != "":
                game["TimeControl"] = "%d+%d" % (
                    self.json_field(api, "clock/initial"),
                    self.json_field(api, "clock/increment"),
                )
            else:
                game["TimeControl"] = "%dd" % (
                    self.json_field(api, "correspondence/increment") // 86400
                )
            game["Result"] = "*"
            game["_moves"] = ""
            moves = self.json_field(api, "steps")
            for move in moves:
                if move["ply"] > 0:
                    game["_moves"] += " %s" % move["san"]
            return self.rebuild_pgn(game)

        # Logic for the studies
        elif self.url_type == TYPE_STUDY:
            url = f"https://lichess.{self.url_tld}/study/{self.id}.pgn"
            return self.download(url, userAgent=True)

        # Logic for the puzzles
        elif self.url_type == TYPE_PUZZLE:
            # The API doesn't provide the history of the moves
            # chessgame = self.query_api('/training/%s/load' % self.id)

            # Fetch the puzzle
            url = f"https://lichess.{self.url_tld}/training/{self.id}"
            page = self.download(url)
            if page is None:
                return None

            # Extract the JSON
            page = page.replace("\n", "")
            pos1 = page.find("lichess.puzzle =")
            if pos1 == -1:
                return None
            pos1 = page.find('"game"', pos1 + 1)
            if pos1 == -1:
                return None
            c = 1
            pos2 = pos1
            while pos2 < len(page):
                pos2 += 1
                if page[pos2] == "{":
                    c += 1
                if page[pos2] == "}":
                    c -= 1
                if c == 0:
                    break
            if c != 0:
                return None

            # Header
            bourne = page[pos1 - 1 : pos2 + 1]
            chessgame = self.json_loads(bourne)
            puzzle = self.json_field(chessgame, "puzzle")
            if puzzle == "":
                return None
            game = {}
            game["_url"] = "https://lichess.{}/{}#{}".format(
                self.url_tld,
                self.json_field(puzzle, "gameId"),
                self.json_field(puzzle, "initialPly"),
            )
            game["Site"] = "lichess.%s" % self.url_tld
            rating = self.json_field(puzzle, "rating")
            game["Event"] = "Puzzle %d, rated %s" % (
                self.json_field(puzzle, "id"),
                rating,
            )
            game["Result"] = "*"
            game["X_ID"] = self.json_field(puzzle, "id")
            game["X_TimeControl"] = self.json_field(chessgame, "game/clock")
            game["X_Rating"] = rating
            game["X_Attempts"] = self.json_field(puzzle, "attempts")
            game["X_Vote"] = self.json_field(puzzle, "vote")

            # Players
            players = self.json_field(chessgame, "game/players")
            if not isinstance(players, list):
                return None
            for p in players:
                if p["color"] == "white":
                    t = "White"
                elif p["color"] == "black":
                    t = "Black"
                else:
                    return None
                pos1 = p["name"].find(" (")
                if pos1 == -1:
                    game[t] = p["name"]
                else:
                    game[t] = p["name"][:pos1]
                    game[t + "Elo"] = p["name"][pos1 + 2 : -1]

            # Moves
            moves = self.json_field(chessgame, "game/treeParts")
            if not isinstance(moves, list):
                return None
            game["_moves"] = ""
            for m in moves:
                if m["ply"] in [0, "0"]:
                    game["SetUp"] = "1"
                    game["FEN"] = m["fen"]
                else:
                    game["_moves"] += "%s " % m["san"]

            # Solution
            game["_moves"] += " {Solution: "
            puzzle = self.json_field(puzzle, "branch")
            while True:
                game["_moves"] += "%s " % self.json_field(puzzle, "san")
                puzzle = self.json_field(puzzle, "children")
                if len(puzzle) == 0:
                    break
                puzzle = puzzle[0]
            game["_moves"] += "}"

            # Rebuild the PGN game
            return self.rebuild_pgn(game)


# ChessGames.com
class InternetGameChessgames(InternetGameInterface):
    def reset(self):
        InternetGameInterface.reset(self)
        self.computer = False

    def get_identity(self):
        return "ChessGames.com", CAT_DL

    def assign_game(self, url):
        # Verify the hostname
        parsed = urlparse(url)
        if parsed.netloc.lower() not in ["www.chessgames.com", "chessgames.com"]:
            return False

        # Read the arguments
        args = parse_qs(parsed.query)
        if "gid" in args:
            gid = args["gid"][0]
            if gid.isdigit() and gid != "0":
                self.id = gid
                self.computer = ("comp" in args) and (args["comp"][0] == "1")
                return True
        return False

    def download_game(self):
        # Check
        if self.id is None:
            return None

        # First try with computer analysis
        url = "http://www.chessgames.com/pgn/pychess.pgn?gid=" + self.id
        if self.computer:
            pgn = self.download(url + "&comp=1")
            if pgn in [None, ""] or "NO SUCH GAME" in pgn:
                self.computer = False
            else:
                return pgn

        # Second try without computer analysis
        if not self.computer:
            pgn = self.download(url)
            if pgn in [None, ""] or "NO SUCH GAME" in pgn:
                return None
            else:
                return pgn


# FicsGames.org
class InternetGameFicsgames(InternetGameInterface):
    def get_identity(self):
        return "FicsGames.org", CAT_DL

    def assign_game(self, url):
        # Verify the URL
        parsed = urlparse(url)
        if (
            parsed.netloc.lower() not in ["www.ficsgames.org", "ficsgames.org"]
            or "show" not in parsed.path.lower()
        ):
            return False

        # Read the arguments
        args = parse_qs(parsed.query)
        if "ID" in args:
            gid = args["ID"][0]
            if gid.isdigit() and gid != "0":
                self.id = gid
                return True
        return False

    def download_game(self):
        # Check
        if self.id is None:
            return None

        # Download
        pgn = self.download(
            "http://ficsgames.org/cgi-bin/show.cgi?ID=%s;action=save" % self.id
        )
        if pgn in [None, ""] or "not found in GGbID" in pgn:
            return None
        else:
            return pgn


# ChessTempo.com
class InternetGameChesstempo(InternetGameInterface):
    def get_identity(self):
        return "ChessTempo.com", CAT_WS

    def assign_game(self, url):
        # Puzzles
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?chesstempo\.com\/chess-tactics\/(\d+)", re.IGNORECASE
        )
        m = rxp.match(url)
        if m is not None:
            gid = str(m.group(2))
            if gid.isdigit() and gid != "0":
                self.id = gid
                self.url_type = TYPE_PUZZLE
                return True

        # Games
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?chesstempo\.com\/gamedb\/game\/(\d+)", re.IGNORECASE
        )
        m = rxp.match(url)
        if m is not None:
            gid = str(m.group(2))
            if gid.isdigit() and gid != "0":
                self.id = gid
                self.url_type = TYPE_GAME
                return True
        return False

    async def download_game(self):
        # Check
        if None in [self.id, self.url_type]:
            return

        # Games
        if self.url_type == TYPE_GAME:
            pgn = self.download(
                "http://chesstempo.com/requests/download_game_pgn.php?gameids=%s"
                % self.id,
                userAgent=True,
            )  # Else a random game is retrieved
            if pgn is not None and len(pgn) > 128:
                self.data = pgn

        # Puzzles
        elif self.url_type == TYPE_PUZZLE:
            # Open a websocket to retrieve the puzzle
            async def coro(self):
                result = None
                ws = await websockets.connect(
                    "wss://chesstempo.com:443/ws",
                    origin="https://chesstempo.com",
                    extra_headers=[("User-agent", self.userAgent)],
                    ping_interval=None,
                )
                try:
                    # Check the welcome message
                    data = await ws.recv()
                    data = self.json_loads(data)
                    if (data["eventName"] == "connectionStarted") and (
                        data["data"] == "started"
                    ):
                        # Call the puzzle
                        await ws.send(
                            '{"eventName":"get-problem-session-data","data":{"problemSetId":1,"sessionSize":20}}'
                        )
                        await ws.send(
                            '{"eventName":"set-problem-difficulty","data":{"difficulty":"","problemSetId":1}}'
                        )
                        await ws.send(
                            '{"eventName":"get-tactic","data":{"problemId":%s,"vo":false}}'
                            % self.id
                        )

                        for i in range(3):
                            data = await ws.recv()
                            data = self.json_loads(data)
                            if data["eventName"] == "get-tactic-result":
                                if data["enc"]:
                                    data = "".join(
                                        map(
                                            lambda v: (
                                                v
                                                if v < "0" or v > "9"
                                                else str((9 + int(v)) % 10)
                                            ),
                                            list(data["data"]),
                                        )
                                    )
                                    result = b64decode(data).decode().strip()
                                else:
                                    result = data["data"].strip()
                                if result == "":
                                    result = None
                finally:
                    await ws.close()
                self.data = result

            await coro(self)
            if self.data is None:
                return

            # Rebuild the puzzle
            puzzle = self.json_loads(self.data)
            game = {}
            game["_url"] = "https://chesstempo.com/chess-tactics/%s" % self.id
            game["Event"] = "Puzzle %s" % self.json_field(
                puzzle, "tacticInfo/problem_id"
            )
            game["White"] = _("White")
            game["Black"] = _("Black")
            game["Result"] = "*"
            game["FEN"] = self.json_field(puzzle, "tacticInfo/startPosition")
            game["SetUp"] = "1"
            game["_moves"] = "{{{}}} {}".format(
                self.json_field(puzzle, "tacticInfo/prevmove"),
                self.json_field(puzzle, "tacticInfo/moves"),
            )
            self.data = self.rebuild_pgn(game)


# 365chess.com
class InternetGame365chess(InternetGameInterface):
    def get_identity(self):
        return "365chess.com", CAT_HTML

    def assign_game(self, url):
        # Verify the URL
        parsed = urlparse(url)
        if parsed.netloc.lower() not in ["www.365chess.com", "365chess.com"]:
            return False
        ppl = parsed.path.lower()
        if ppl == "/game.php":
            key = "gid"
        elif ppl == "/view_game.php":
            key = "g"
        else:
            return False

        # Read the arguments
        args = parse_qs(parsed.query)
        if key in args:
            gid = args[key][0]
            if gid.isdigit() and gid != "0":
                self.id = gid
                return True
        return False

    def download_game(self):
        # Download
        if self.id is None:
            return None
        url = "https://www.365chess.com/game.php?gid=%s" % self.id
        page = self.download(url)
        if page is None:
            return None

        # Played moves
        game = {}
        pos1 = page.find("chess_game.Init({")
        pos1 = page.find(",pgn:'", pos1)
        pos2 = page.find("'", pos1 + 6)
        if -1 in [pos1, pos2]:
            return None
        game["_moves"] = page[pos1 + 6 : pos2]

        # Result
        result = game["_moves"].split(" ")[-1]
        if result in reprResult:
            game["Result"] = result
            game["_moves"] = " ".join(game["_moves"].split(" ")[0:-1])

        # Header
        game["_url"] = url
        lines = page.replace("<td", "\n<td").split("\n")
        rxp = re.compile(
            r"^([\w\-,\s]+)(\(([0-9]+)\))? vs\. ([\w\-,\s]+)(\(([0-9]+)\))?$",
            re.IGNORECASE,
        )
        game["White"] = _("Unknown")
        game["Black"] = _("Unknown")
        for line in lines:
            line = line.strip()

            # Event
            for tag in ["Event", "Site", "Date", "Round"]:
                if tag + ":" in line:
                    pos1 = line.find(tag + ":")
                    pos1 = line.find(" ", pos1)
                    pos2 = line.find("<", pos1)
                    if -1 not in [pos1, pos2]:
                        v = line[pos1 + 1 : pos2]
                        if tag == "Date":
                            v = "{}.{}.{}".format(
                                v[-4:], v[:2], v[3:5]
                            )  # mm/dd/yyyy --> yyyy.mm.dd
                        game[tag] = v

            # Players
            line = self.strip_html(line).strip()
            m = rxp.match(line)
            if m is not None:
                game["White"] = str(m.group(1)).strip()
                if m.group(3) is not None:
                    game["WhiteElo"] = str(m.group(3)).strip()
                game["Black"] = str(m.group(4)).strip()
                if m.group(6) is not None:
                    game["BlackElo"] = str(m.group(6)).strip()

        # Rebuild the PGN game
        return self.rebuild_pgn(game)


# ChessPastebin.com
class InternetGameChesspastebin(InternetGameInterface):
    def get_identity(self):
        return "ChessPastebin.com", CAT_HTML

    def assign_game(self, url):
        return self.reacts_to(url, "chesspastebin.com")

    def download_game(self):
        # Download
        if self.id is None:
            return None
        page = self.download(self.id)
        if page is None:
            return None

        # Extract the game ID
        rxp = re.compile(
            r".*?\<div id=\"([0-9]+)_board\"\>\<\/div\>.*?", flags=re.IGNORECASE
        )
        m = rxp.match(page.replace("\n", ""))
        if m is None:
            return None
        gid = m.group(1)

        # Definition of the parser
        class chesspastebinparser(HTMLParser):
            def __init__(self):
                HTMLParser.__init__(self)
                self.tag_ok = False
                self.pgn = None

            def handle_starttag(self, tag, attrs):
                if tag.lower() == "div":
                    for k, v in attrs:
                        if k.lower() == "id" and v == gid:
                            self.tag_ok = True

            def handle_data(self, data):
                if self.pgn is None and self.tag_ok:
                    self.pgn = data

        # Read the PGN
        parser = chesspastebinparser()
        parser.feed(page)
        pgn = parser.pgn
        if (
            pgn is not None
        ):  # Any game must start with '[' to be considered further as valid
            pgn = pgn.strip()
            if not pgn.startswith("["):
                pgn = '[Annotator "ChessPastebin.com"]\n%s' % pgn
        return pgn


# TheChessWorld.com
class InternetGameThechessworld(InternetGameInterface):
    def get_identity(self):
        return "TheChessWorld.com", CAT_DL

    def assign_game(self, url):
        return self.reacts_to(url, "thechessworld.com")

    def download_game(self):
        # Check
        if self.id is None:
            return None

        # Find the links
        links = []
        if self.id.lower().endswith(".pgn"):
            links.append(self.id)
        else:
            # Download the page
            data = self.download(self.id)
            if data is None:
                return None

            # Finds the games
            rxp = re.compile(r".*pgn_uri:.*'([^']+)'.*", re.IGNORECASE)
            lines = data.split("\n")
            for line in lines:
                m = rxp.match(line)
                if m is not None:
                    links.append("https://www.thechessworld.com" + m.group(1))

        # Collect the games
        return self.download_list(links)


# Chess.org
class InternetGameChessOrg(InternetGameInterface):
    def get_identity(self):
        return "Chess.org", CAT_WS

    def assign_game(self, url):
        rxp = re.compile(
            r"^https?:\/\/chess\.org\/play\/([a-f0-9\-]+)[\/\?\#]?", re.IGNORECASE
        )
        m = rxp.match(url)
        if m is not None:
            id = str(m.group(1))
            if len(id) == 36:
                self.id = id
                return True
        return False

    async def download_game(self):
        # Check
        if self.id is None:
            return

        # Fetch the page to retrieve the encrypted user name
        url = "https://chess.org/play/%s" % self.id
        page = self.download(url)
        if page is None:
            return
        lines = page.split("\n")
        name = ""
        for line in lines:
            pos1 = line.find("encryptedUsername")
            if pos1 != -1:
                pos1 = line.find("'", pos1)
                pos2 = line.find("'", pos1 + 1)
                if pos2 > pos1:
                    name = line[pos1 + 1 : pos2]
                    break
        if name == "":
            return

        # Random elements to get a unique URL
        rndI = randint(1, 1000)
        rndS = "".join(choice(string.ascii_lowercase) for i in range(8))

        # Open a websocket to retrieve the chess data
        async def coro(self):
            url = "wss://chess.org:443/play-sockjs/%d/%s/websocket" % (rndI, rndS)
            log.debug("Websocket connecting to %s" % url)
            ws = await websockets.connect(
                url, origin="https://chess.org:443", ping_interval=None
            )
            try:
                # Server: Hello
                data = await ws.recv()
                if data == "o":  # Open
                    # Client: I am XXX, please open the game YYY
                    await ws.send(f'["{name} {self.id}"]')
                    data = await ws.recv()

                    # Server: some data
                    if data[:1] == "a":
                        data = data[3:-2]
                        if data not in [None, ""]:
                            self.data = data
            finally:
                await ws.close()

        await coro(self)
        if self.data is None:
            return

        # Parses the game
        chessgame = self.json_loads(self.data.replace('\\"', '"'))
        game = {}
        game["_url"] = url
        board = LBoard(variant=FISCHERRANDOMCHESS)

        # Player info
        if self.json_field(chessgame, "creatorColor") == "1":  # White=1, Black=0
            creator = "White"
            opponent = "Black"
        else:
            creator = "Black"
            opponent = "White"
        game[creator] = self.json_field(chessgame, "creatorId")
        elo = self.json_field(chessgame, "creatorPoint")
        if elo not in ["", "0", 0]:
            game[creator + "Elo"] = elo
        game[opponent] = self.json_field(chessgame, "opponentId")
        elo = self.json_field(chessgame, "opponentPoint")
        if elo not in ["", "0", 0]:
            game[opponent + "Elo"] = elo

        # Game info
        startPos = self.json_field(chessgame, "startPos")
        if startPos not in ["", "startpos"]:
            game["SetUp"] = "1"
            game["FEN"] = startPos
            game["Variant"] = CHESS960
            try:
                board.applyFen(startPos)
            except Exception:
                self.data = None
                return
        else:
            board.applyFen("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w AHah - 0 1")
        time = self.json_field(chessgame, "timeLimitSecs")
        inc = self.json_field(chessgame, "timeBonusSecs")
        if "" not in [time, inc]:
            game["TimeControl"] = f"{time}+{inc}"
        resultTable = [
            (0, "*", "Game started"),  # ALIVE
            (1, "1-0", "White checkmated"),  # WHITE_MATE
            (2, "0-1", "Black checkmated"),  # BLACK_MATE
            (3, "1/2-1/2", "White stalemated"),  # WHITE_STALEMATE
            (4, "1/2-1/2", "Black stalemated"),  # BLACK_STALEMATE
            (5, "1/2-1/2", "Insufficient material"),  # DRAW_NO_MATE
            (6, "1/2-1/2", "50-move rule"),  # DRAW_50
            (7, "1/2-1/2", "Threefold repetition"),  # DRAW_REP
            (8, "1/2-1/2", "Mutual agreement"),  # DRAW_AGREE
            (9, "0-1", "White resigned"),  # WHITE_RESIGN
            (10, "1-0", "Black resigned"),  # BLACK_RESIGN
            (11, "0-1", "White canceled"),  # WHITE_CANCEL
            (12, "1-0", "Black canceled"),  # BLACK_CANCEL
            (13, "1-0", "White out of time"),  # WHITE_NO_TIME
            (14, "0-1", "Black out of time"),  # BLACK_NO_TIME
            (15, "*", "Not started"),
        ]  # NOT_STARTED
        state = self.json_field(chessgame, "state")
        result = "*"
        reason = "Unknown reason %d" % state
        for rtID, rtScore, rtMsg in resultTable:
            if rtID == state:
                result = rtScore
                reason = rtMsg
                break
        game["Result"] = result
        game["_reason"] = reason

        # Moves
        game["_moves"] = ""
        moves = self.json_field(chessgame, "lans")
        if moves == "":
            self.data = None
            return
        moves = moves.split(" ")
        for move in moves:
            try:
                if self.use_an:
                    move = parseAny(board, move)
                    game["_moves"] += toSAN(board, move) + " "
                    board.applyMove(move)
                else:
                    game["_moves"] += move + " "
            except Exception:
                self.data = None
                return

        # Rebuild the PGN game
        self.data = self.rebuild_pgn(game)


# Europe-Echecs.com
class InternetGameEuropeechecs(InternetGameInterface):
    def get_identity(self):
        return "Europe-Echecs.com", CAT_DL

    def assign_game(self, url):
        return self.reacts_to(url, "europe-echecs.com")

    def download_game(self):
        # Check
        if self.id is None:
            return None

        # Find the links
        links = []
        if self.id.lower().endswith(".pgn"):
            links.append(self.id)
        else:
            # Download the page
            page = self.download(self.id)
            if page is None:
                return None

            # Find the chess widgets
            rxp = re.compile(
                r".*class=\"cbwidget\"\s+id=\"([0-9a-f]+)_container\".*", re.IGNORECASE
            )
            lines = page.split("\n")
            for line in lines:
                m = rxp.match(line)
                if m is not None:
                    links.append(
                        "https://www.europe-echecs.com/embed/doc_%s.pgn" % m.group(1)
                    )

        # Collect the games
        return self.download_list(links)


# GameKnot.com
class InternetGameGameknot(InternetGameInterface):
    def get_identity(self):
        return "GameKnot.com", CAT_HTML

    def assign_game(self, url):
        # Verify the hostname
        parsed = urlparse(url.lower())
        if parsed.netloc not in ["www.gameknot.com", "gameknot.com"]:
            return False

        # Verify the page
        if parsed.path == "/analyze-board.pl":
            ttype = TYPE_GAME
            tkey = "bd"
        elif parsed.path == "/chess-puzzle.pl":
            ttype = TYPE_PUZZLE
            tkey = "pz"
        else:
            return False

        # Read the arguments
        args = parse_qs(parsed.query)
        if tkey in args:
            gid = args[tkey][0]
            if gid.isdigit() and gid != "0":
                self.id = gid
                self.url_type = ttype
                return True
        return False

    def download_game(self):
        # Check
        if self.url_type not in [TYPE_GAME, TYPE_PUZZLE] or self.id is None:
            return None

        # Download
        if self.url_type == TYPE_GAME:
            url = "https://gameknot.com/analyze-board.pl?bd=%s" % self.id
        elif self.url_type == TYPE_PUZZLE:
            url = "https://gameknot.com/chess-puzzle.pl?pz=%s" % self.id
        page = self.download(url, userAgent=True)
        if page is None:
            return None

        # Library
        def extract_variables(page, structure):
            game = {}
            for var, type, tag in structure:
                game[tag] = ""
            lines = page.split(";")
            for line in lines:
                for var, type, tag in structure:
                    pos1 = line.find(var)
                    if pos1 == -1:
                        continue
                    if type == "s":
                        pos1 = line.find("'", pos1 + 1)
                        pos2 = line.find("'", pos1 + 1)
                        if pos2 > pos1:
                            game[tag] = line[pos1 + 1 : pos2]
                    elif type == "i":
                        pos1 = line.find("=", pos1 + 1)
                        if pos1 != -1:
                            txt = line[pos1 + 1 :].strip()
                            if txt not in ["", "0"]:
                                game[tag] = txt
                    else:
                        assert False
            return game

        # Logic for the puzzles
        if self.url_type == TYPE_PUZZLE:
            structure = [
                ("puzzle_id", "i", "_id"),
                ("puzzle_fen", "s", "FEN"),
                ("load_solution(", "s", "_solution"),
            ]
            game = extract_variables(page, structure)
            game["_url"] = "https://gameknot.com/chess-puzzle.pl?pz=%s" % game["_id"]
            game["White"] = _("White")
            game["Black"] = _("Black")
            game["Result"] = "*"
            if game["FEN"] != "":
                game["SetUp"] = "1"
            if game["_solution"] != "":
                list = game["_solution"].split("|")
                game["_moves"] = " {Solution:"
                nextid = "0"
                for item in list:
                    item = item.split(",")
                    # 0 = identifier of the move
                    # 1 = player
                    # 2 = identifier of the previous move
                    # 3 = count of following moves
                    # 4 = algebraic notation of the move
                    # 5 = UCI notation of the move
                    # 6 = ?
                    # 7 = identifier of the next move
                    # > = additional moves for the current line
                    curid = item[0]
                    if curid != nextid:
                        continue
                    if len(item) == 4:
                        break
                    nextid = item[7]
                    if self.use_an:
                        move = item[4]
                    else:
                        move = item[5]
                    game["_moves"] += " %s" % move
                game["_moves"] += "}"

        # Logic for the games
        elif self.url_type == TYPE_GAME:
            # Header
            structure = [
                ("anbd_movelist", "s", "_moves"),
                ("anbd_result", "i", "Result"),
                ("anbd_player_w", "s", "White"),
                ("anbd_player_b", "s", "Black"),
                ("anbd_rating_w", "i", "WhiteElo"),
                ("anbd_rating_b", "i", "BlackElo"),
                ("anbd_title", "s", "Event"),
                ("anbd_timestamp", "s", "Date"),
                ("export_web_input_result_text", "s", "_reason"),
            ]
            game = extract_variables(page, structure)
            if game["Result"] == "1":
                game["Result"] = "1-0"
            elif game["Result"] == "2":
                game["Result"] = "1/2-1/2"
            elif game["Result"] == "3":
                game["Result"] = "0-1"
            else:
                game["Result"] = "*"

            # Body
            board = LBoard()
            board.applyFen(FEN_START)
            moves = game["_moves"].split("-")
            game["_moves"] = ""
            for move in moves:
                if move == "":
                    break
                try:
                    if self.use_an:
                        kmove = parseAny(board, move)
                        game["_moves"] += toSAN(board, kmove) + " "
                        board.applyMove(kmove)
                    else:
                        game["_moves"] += move + " "
                except Exception:
                    return None

        # Rebuild the PGN game
        return unquote(self.rebuild_pgn(game))


# Chess.com
class InternetGameChessCom(InternetGameInterface):
    def get_identity(self):
        return "Chess.com", CAT_MISC

    def assign_game(self, url):
        # Positions
        parsed = urlparse(url)
        if parsed.netloc.lower() in ["www.chess.com", "chess.com"]:
            args = parse_qs(parsed.query)
            if "fen" in args:
                fen = args["fen"][0]
                if self.is_fen(fen):
                    self.url_type = TYPE_FEN
                    self.id = fen
                    return True

        # Puzzles
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?chess\.com\/([a-z\/]+)?(puzzles)\/problem\/([0-9]+)[\/\?\#]?",
            re.IGNORECASE,
        )
        m = rxp.match(url)
        if m is not None:
            self.url_type = m.group(3).lower()
            self.id = m.group(4)
            return True

        # Games
        url = url.replace("/live#g=", "/live/game/").replace(
            "/daily#g=", "/daily/game/"
        )
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?chess\.com\/([a-z\/]+)?(live|daily)\/([a-z\/]+)?([0-9]+)[\/\?\#]?",
            re.IGNORECASE,
        )
        m = rxp.match(url)
        if m is not None:
            self.url_type = m.group(3).lower()
            self.id = m.group(5)
            return True
        return False

    def decode_move(self, move):
        # Mapping
        map = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!?{~}(^)[_]@#$,./&-*++="
        pieces = "qnrbkp"

        # Analyze the move
        sFrom = sTo = sPromo = sDrop = ""
        posFrom = map.index(move[:1])
        posTo = map.index(move[1:])
        if posTo > 63:
            sPromo = pieces[(posTo - 64) // 3]
            posTo = posFrom + (-8 if posFrom < 16 else 8) + (posTo - 1) % 3 - 1
        if posFrom > 75:
            sDrop = pieces[posFrom - 79].upper() + "@"
        else:
            sFrom = map[posFrom % 8] + str(posFrom // 8 + 1)
        sTo = map[posTo % 8] + str(posTo // 8 + 1)
        return f"{sDrop}{sFrom}{sTo}{sPromo}"

    def download_game(self):
        # Check
        if None in [self.id, self.url_type]:
            return None

        # Positions
        if self.url_type == TYPE_FEN:
            return '[Site "chess.com"]\n[White "{}"]\n[Black "{}"]\n[SetUp "1"]\n[FEN "{}"]\n\n*'.format(
                _("White"), _("Black"), self.id
            )

        # Puzzles
        elif self.url_type == "puzzles":
            url = "https://www.chess.com/puzzles/problem/%s" % self.id
            page = self.download(url, userAgent=True)  # Else 403 Forbidden
            if page is None:
                return None

            # Definition of the parser
            class chesscomparser(HTMLParser):
                def __init__(self):
                    HTMLParser.__init__(self)
                    self.pgn = None

                def handle_starttag(self, tag, attrs):
                    if tag.lower() == "div":
                        for k, v in attrs:
                            if k.lower() == "data-puzzle":
                                self.pgn = v.replace("&quote;", '"').replace("\\/", "/")

            # Parse the page
            parser = chesscomparser()
            parser.feed(page)
            if parser.pgn is None:
                return None

            # Load the JSON
            puzzle = self.json_loads(parser.pgn)
            if puzzle in [None, ""]:
                return None

            # Get the game directly if available
            pgn = self.json_field("puzzle", "pgn")
            if pgn != "":
                pgn = pgn.replace("\\n", "\n")
                pgn = pgn.replace('\\"', '"')
                pgn = pgn.replace("     ", " ")
                pgn = pgn.replace(". ...", "...")
                pgn = pgn.replace("\\t", "    ")
                return pgn

            # Rebuild the puzzle
            game = {}
            game["_url"] = url
            refid = self.json_field("puzzle", "gameLiveId")
            if refid not in [None, 0, ""]:
                game["_url"] = "https://www.chess.com/live/game/%d" % refid
            else:
                refid = self.json_field("puzzle", "gameId")
                if refid not in [None, 0, ""]:
                    game["_url"] = "https://www.chess.com/daily/game/%d" % refid
            rating = self.json_field(puzzle, "rating")
            game["Event"] = "Puzzle {}, rated {}".format(
                str(self.json_field(puzzle, "id")), rating
            )
            game["White"] = _("White")
            game["Black"] = _("Black")
            game["Result"] = "*"
            game["TimeControl"] = "%d+0" % self.json_field(puzzle, "averageSeconds")
            game["FEN"] = self.json_field(puzzle, "initialFen")
            game["SetUp"] = "1"
            game["X_ID"] = self.json_field(puzzle, "id")
            game["X_Rating"] = rating
            game["X_Attempts"] = self.json_field(puzzle, "attemptCount")
            game["X_PassRate"] = self.json_field(puzzle, "passRate")
            buffer = self.json_field(puzzle, "internalNote").strip()
            if buffer == "":
                buffer = "{%s}" % _("The first move is not provided")
            game["_moves"] = buffer

        # Games
        else:
            if self.url_type == "live":
                # API since October 2020
                url = "https://www.chess.com/callback/live/game/%s" % self.id
                bourne = self.send_xhr(url, {}, userAgent=True)
                if bourne is None:
                    return None
            else:
                # Fetch the page
                url = f"https://www.chess.com/{self.url_type}/game/{self.id}"
                page = self.download(url, userAgent=True)
                if page is None:
                    return None

                # Extract the JSON
                bourne = ""
                pos1 = page.find("window.chesscom.dailyGame")
                if pos1 != -1:
                    pos1 = page.find("(", pos1)
                    pos2 = page.find(")", pos1 + 1)
                    if pos2 > pos1:
                        bourne = (
                            page[pos1 + 2 : pos2 - 1]
                            .replace("\\\\\\/", "/")
                            .replace('\\"', '"')
                        )
                if bourne == "":
                    return None

            # Read the JSON
            chessgame = self.json_loads(bourne)
            if (
                not self.allow_extra
                and self.json_field(chessgame, "game/isRated")
                and not self.json_field(chessgame, "game/isFinished")
            ):
                return None
            chessgame = self.json_field(chessgame, "game")
            if chessgame == "":
                return None

            # Header
            headers = self.json_field(chessgame, "pgnHeaders")
            if headers == "":
                game = {}
            else:
                game = headers
            if "Variant" in game and game["Variant"] == "Chess960":
                game["Variant"] = CHESS960
            game["_url"] = url.replace("/callback/", "/")

            # Body
            moves = self.json_field(chessgame, "moveList")
            if moves == "":
                return None
            game["_moves"] = ""
            if "Variant" in game and game["Variant"] == "Crazyhouse":
                board = LBoard(variant=CRAZYHOUSECHESS)
            else:
                board = LBoard(variant=FISCHERRANDOMCHESS)
            if "FEN" in game:
                board.applyFen(game["FEN"])
            else:
                board.applyFen(FEN_START)
            while len(moves) > 0:
                move = self.decode_move(moves[:2])
                moves = moves[2:]
                try:
                    if self.use_an:
                        kmove = parseAny(board, move)
                        game["_moves"] += toSAN(board, kmove) + " "
                        board.applyMove(kmove)
                    else:
                        game["_moves"] += move + " "
                except Exception:
                    return None

        # Final PGN
        return self.rebuild_pgn(game)


# Schach-Spielen.eu
class InternetGameSchachspielen(InternetGameInterface):
    def get_identity(self):
        return "Schach-Spielen.eu", CAT_HTML

    def assign_game(self, url):
        rxp = re.compile(
            r"^https?:\/\/(www\.)?schach-spielen\.eu\/(game|analyse)\/([a-z0-9]+)[\/\?\#]?",
            re.IGNORECASE,
        )
        m = rxp.match(url)
        if m is not None:
            gid = m.group(3)
            if len(gid) == 8:
                self.id = gid
                return True
        return False

    def download_game(self):
        # Download
        if self.id is None:
            return None
        page = self.download("https://www.schach-spielen.eu/analyse/%s" % self.id)
        if page is None:
            return None

        # Definition of the parser
        class schachspielenparser(HTMLParser):
            def __init__(self):
                HTMLParser.__init__(self)
                self.tag_ok = False
                self.pgn = None

            def handle_starttag(self, tag, attrs):
                if tag.lower() == "textarea":
                    for k, v in attrs:
                        if k.lower() == "id" and v == "pgnText":
                            self.tag_ok = True

            def handle_data(self, data):
                if self.pgn is None and self.tag_ok:
                    self.pgn = data

        # Read the PGN
        parser = schachspielenparser()
        parser.feed(page)
        pgn = parser.pgn
        if pgn is not None:
            pgn = pgn.replace('[Variant "chess960"]', '[Variant "%s"]' % CHESS960)
        return pgn


# RedHotPawn.com
class InternetGameRedhotpawn(InternetGameInterface):
    def get_identity(self):
        return "RedHotPawn.com", CAT_HTML

    def assign_game(self, url):
        # Verify the URL
        parsed = urlparse(url)
        if parsed.netloc.lower() not in ["www.redhotpawn.com", "redhotpawn.com"]:
            return False

        # Verify the path
        ppl = parsed.path.lower()
        if "chess-game-" in ppl:
            ttype = TYPE_GAME
            key = "gameid"
        elif "chess-puzzle-" in ppl:
            ttype = TYPE_PUZZLE
            if "chess-puzzle-serve" in url.lower():
                self.url_type = ttype
                self.id = url
                return True
            else:
                key = "puzzleid"
        else:
            return False

        # Read the arguments
        args = parse_qs(parsed.query)
        if key in args:
            gid = args[key][0]
            if gid.isdigit() and gid != "0":
                self.url_type = ttype
                self.id = gid
                return True
        return False

    def download_game(self):
        # Download
        if self.id is None:
            return None
        if self.url_type == TYPE_GAME:
            url = (
                "https://www.redhotpawn.com/pagelet/view/game-pgn.php?gameid=%s"
                % self.id
            )
        elif self.url_type == TYPE_PUZZLE:
            if "://" in self.id:
                url = self.id
                event = _("Puzzle")
            else:
                url = (
                    "https://www.redhotpawn.com/chess-puzzles/chess-puzzle-solve.php?puzzleid=%s"
                    % self.id
                )
                event = _("Puzzle %s") % self.id
        else:
            return None
        page = self.download(url)
        if page is None:
            return None

        # Logic for the games
        if self.url_type == TYPE_GAME:
            # Parser
            class redhotpawnparser(HTMLParser):
                def __init__(self):
                    HTMLParser.__init__(self)
                    self.tag_ok = False
                    self.pgn = None

                def handle_starttag(self, tag, attrs):
                    if tag.lower() == "textarea":
                        self.tag_ok = True

                def handle_data(self, data):
                    if self.pgn is None and self.tag_ok:
                        self.pgn = data

            # Extractor
            parser = redhotpawnparser()
            parser.feed(page)
            return parser.pgn.strip()

        # Logic for the puzzles
        elif self.url_type == TYPE_PUZZLE:
            pos1 = page.find("var g_startFenStr")
            if pos1 != -1:
                pos1 = page.find("'", pos1)
                pos2 = page.find("'", pos1 + 1)
                if pos2 > pos1:
                    game = {}
                    game["_url"] = url
                    game["FEN"] = page[pos1 + 1 : pos2]
                    game["SetUp"] = "1"
                    game["Event"] = event
                    game["White"] = _("White")
                    game["Black"] = _("Black")
                    pos1 = page.find("<h4>")
                    pos2 = page.find("</h4>", pos1)
                    if pos1 != -1 and pos2 > pos1:
                        game["_moves"] = "{%s}" % page[pos1 + 4 : pos2]
                    return self.rebuild_pgn(game)

        return None


# Chess-Samara.ru
class InternetGameChesssamara(InternetGameInterface):
    def get_identity(self):
        return "Chess-Samara.ru", CAT_DL

    def assign_game(self, url):
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?chess-samara\.ru\/(\d+)\-", re.IGNORECASE
        )
        m = rxp.match(url)
        if m is not None:
            gid = str(m.group(2))
            if gid.isdigit() and gid != "0":
                self.id = gid
                return True
        return False

    def download_game(self):
        # Check
        if self.id is None:
            return None

        # Download
        pgn = self.download("https://chess-samara.ru/view/pgn.html?gameid=%s" % self.id)
        if pgn is None or len(pgn) == 0:
            return None
        else:
            return pgn


# 2700chess.com
class InternetGame2700chess(InternetGameInterface):
    def get_identity(self):
        return "2700chess.com", CAT_HTML

    def assign_game(self, url):
        # Verify the hostname
        parsed = urlparse(url)
        if parsed.netloc.lower() not in ["www.2700chess.com", "2700chess.com"]:
            return False

        # Refactor the direct link
        if parsed.path.lower() == "/games/download":
            args = parse_qs(parsed.query)
            if "slug" in args:
                self.id = "https://2700chess.com/games/%s" % args["slug"][0]
                return True

        # Verify the path
        if parsed.path.startswith("/games/"):
            self.id = url
            return True
        else:
            return False

    def download_game(self):
        # Download
        if self.id is None:
            return None
        page = self.download(self.id)
        if page is None:
            return None

        # Extract the PGN
        lines = page.split(";")
        for line in lines:
            if "analysis.setPgn(" in line:
                pos1 = line.find('"')
                if pos1 != -1:
                    pos2 = pos1
                    while pos2 < len(line):
                        pos2 += 1
                        if line[pos2] == '"' and line[pos2 - 1 : pos2 + 1] != '\\"':
                            pgn = line[pos1 + 1 : pos2]
                            return (
                                pgn.replace('\\"', '"')
                                .replace("\\/", "/")
                                .replace("\\n", "\n")
                                .strip()
                            )
        return None


# Iccf.com
class InternetGameIccf(InternetGameInterface):
    def get_identity(self):
        return "Iccf.com", CAT_DL

    def assign_game(self, url):
        # Verify the hostname
        parsed = urlparse(url)
        if parsed.netloc.lower() not in ["www.iccf.com", "iccf.com"]:
            return False

        # Verify the path
        ppl = parsed.path.lower()
        if "/game" in ppl:
            ttyp = TYPE_GAME
        elif "/event" in ppl:
            ttyp = TYPE_EVENT
        else:
            return False

        # Read the arguments
        args = parse_qs(parsed.query)
        if "id" in args:
            gid = args["id"][0]
            if gid.isdigit() and gid != "0":
                self.url_type = ttyp
                self.id = gid
                return True
        return False

    def download_game(self):
        # Check
        if self.url_type not in [TYPE_GAME, TYPE_EVENT] or self.id is None:
            return None

        # Download
        if self.url_type == TYPE_GAME:
            url = "https://www.iccf.com/GetPGN.aspx?id=%s"
        elif self.url_type == TYPE_EVENT:
            url = "https://www.iccf.com/GetEventPGN.aspx?id=%s"
        pgn = self.download(url % self.id)
        if pgn in [None, ""] or "does not exist." in pgn or "Invalid event" in pgn:
            return None
        else:
            return pgn


# SchachArena.de
class InternetGameSchacharena(InternetGameInterface):
    def get_identity(self):
        return "SchachArena.de", CAT_HTML

    def assign_game(self, url):
        # Verify the URL
        parsed = urlparse(url)
        if (
            parsed.netloc.lower() not in ["www.schacharena.de", "schacharena.de"]
            or "verlauf" not in parsed.path.lower()
        ):
            return False

        # Read the arguments
        args = parse_qs(parsed.query)
        if "brett" in args:
            gid = args["brett"][0]
            if gid.isdigit() and gid != "0":
                self.id = gid
                return True
        return False

    def download_game(self):
        # Check
        if self.id is None:
            return None

        # Download page
        page = self.download(
            "https://www.schacharena.de/new/verlauf.php?brett=%s" % self.id
        )
        if page is None:
            return None

        # Init
        rxp_player = re.compile(
            r".*spielerstatistik.*name=(\w+).*\[([0-9]+)\].*", re.IGNORECASE
        )
        rxp_move = re.compile(r".*<span.*onMouseOut.*fan\(([0-9]+)\).*", re.IGNORECASE)
        rxp_result = re.compile(
            r".*>(1\-0|0\-1|1\/2\-1\/2)\s([^\<]+)<.*", re.IGNORECASE
        )
        player_count = 0
        board = LBoard()
        board.applyFen(FEN_START)

        # Parse
        game = {}
        game["Result"] = "*"
        reason = ""
        game["_moves"] = ""
        game["_url"] = (
            "https://www.schacharena.de/new/verlauf_to_pgn_n.php?brett=%s" % self.id
        )  # If one want to get the full PGN
        lines = page.split("\n")
        for line in lines:
            # Player
            m = rxp_player.match(line)
            if m is not None:
                player_count += 1
                if player_count == 1:
                    tag = "White"
                elif player_count == 2:
                    tag = "Black"
                else:
                    return None
                game[tag] = m.group(1)
                game[tag + "Elo"] = m.group(2)
                continue

            # Move
            m = rxp_move.match(line)
            if m is not None:
                move = m.group(1)
                move = (
                    "_abcdefgh"[int(move[0])]
                    + move[1]
                    + "_abcdefgh"[int(move[2])]
                    + move[3]
                )
                if self.use_an:
                    kmove = parseAny(board, move)
                    move = toSAN(board, kmove)
                    board.applyMove(kmove)
                game["_moves"] += "%s " % move
                continue

            # Result
            m = rxp_result.match(line)
            if m is not None:
                game["Result"] = m.group(1)
                reason = unescape(m.group(2))
                continue

        # Final PGN
        if reason != "":
            game["_moves"] += " {%s}" % reason
        return self.rebuild_pgn(game)


# ChessPuzzle.net
class InternetGameChesspuzzle(InternetGameInterface):
    def get_identity(self):
        return "ChessPuzzle.net", CAT_HTML

    def assign_game(self, url):
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?chesspuzzle\.net\/(Puzzle|Solution)\/([0-9]+)[\/\?\#]?",
            re.IGNORECASE,
        )
        m = rxp.match(url)
        if m is not None:
            gid = str(m.group(3))
            if gid.isdigit() and gid != "0":
                self.id = gid
                return True
        return False

    def download_game(self):
        # Check
        if self.id is None:
            return None

        # Download the puzzle
        page = self.download(
            "https://chesspuzzle.net/Solution/%s" % self.id, userAgent=True
        )  # Else 403 Forbidden
        if page is None:
            return None

        # Definition of the parser
        class chesspuzzleparser(HTMLParser):
            def __init__(self):
                HTMLParser.__init__(self)
                self.last_tag = None
                self.pgn = None

            def handle_starttag(self, tag, attrs):
                self.last_tag = tag.lower()

            def handle_data(self, data):
                if self.pgn is None and self.last_tag == "script":
                    lines = data.split("\n")
                    for line in lines:
                        pos1 = line.find("pgn_text")
                        if pos1 != -1:
                            pos1 = line.find("'", pos1 + 1)
                            pos2 = line.find("'", pos1 + 1)
                            if pos1 != -1 and pos2 > pos1:
                                self.pgn = (
                                    line[pos1 + 1 : pos2]
                                    .replace("]  ", "]\n\n")
                                    .replace("] ", "]\n")
                                    .strip()
                                )
                                break

        # Get the puzzle
        parser = chesspuzzleparser()
        parser.feed(page)
        return parser.pgn


# ChessKing.com
class InternetGameChessking(InternetGameInterface):
    def get_identity(self):
        return "ChessKing.com", CAT_DL

    def assign_game(self, url):
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?chessking\.com\/games\/(ff\/)?([0-9]+)[\/\?\#]?",
            re.IGNORECASE,
        )
        m = rxp.match(url)
        if m is not None:
            gid = str(m.group(3))
            if gid.isdigit() and gid != "0" and len(gid) <= 9:
                if m.group(2) == "ff/":
                    self.url_type = "f"
                else:
                    self.url_type = "g"
                self.id = gid
                return True
        return False

    def download_game(self):
        # Check
        if None in [self.url_type, self.id]:
            return None

        # Download
        id = self.id
        while len(id) < 9:
            id = "0%s" % id
        url = "https://c1.chessking.com/pgn/{}/{}/{}/{}{}.pgn".format(
            self.url_type, id[:3], id[3:6], self.url_type, id
        )
        return self.download(url)


# IdeaChess.com
class InternetGameIdeachess(InternetGameInterface):
    def get_identity(self):
        return "IdeaChess.com", CAT_API

    def assign_game(self, url):
        # Game ID
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?ideachess\.com\/.*\/.*\/([0-9]+)[\/\?\#]?",
            re.IGNORECASE,
        )
        m = rxp.match(url)
        if m is not None:
            gid = str(m.group(2))
            if gid.isdigit() and gid != "0":
                # Game type
                classification = [
                    ("/chess_tactics_puzzles/checkmate_n/", "m"),
                    ("/echecs_tactiques/mat_n/", "m"),
                    ("/scacchi_tattica/scacco_matto_n/", "m"),
                    ("/chess_tactics_puzzles/tactics_n/", "t"),
                    ("/echecs_tactiques/tactiques_n/", "t"),
                    ("/scacchi_tattica/tattica_n/", "t"),
                ]
                for path, ttyp in classification:
                    if path in url.lower():
                        self.url_type = ttyp
                        self.id = gid
                        return True
        return False

    def download_game(self):
        # Check
        if self.url_type is None or self.id is None:
            return None

        # Fetch the puzzle
        api = "http://www.ideachess.com/com/ajax2"
        data = {
            "message": '{{"action":100,"data":{{"problemNumber":{},"kind":"{}"}}}}'.format(
                self.id, self.url_type
            )
        }
        bourne = self.send_xhr(api, data, userAgent=True)
        chessgame = self.json_loads(bourne)
        if self.json_field(chessgame, "action") != 200:
            return None

        # Build the PGN
        game = {}
        if self.url_type == "m":
            game["_url"] = (
                "http://www.ideachess.com/chess_tactics_puzzles/checkmate_n/%s"
                % self.id
            )
        elif self.url_type == "t":
            game["_url"] = (
                "http://www.ideachess.com/chess_tactics_puzzles/tactics_n/%s" % self.id
            )
        else:
            assert False
        game["FEN"] = b64decode(self.json_field(chessgame, "data/FEN")).decode().strip()
        game["SetUp"] = "1"
        game["_moves"] = self.json_field(chessgame, "data/PGN")
        v = self.json_field(chessgame, "data/requiredMoves")
        if v > 0:
            game["Site"] = _("%d moves to find") % v
        list = self.json_field(chessgame, "data/extraInfo").split("|")
        if len(list) == 4:
            game["Event"] = list[0][list[0].find(" ") + 1 :].strip()
            game["Date"] = list[1].strip()
            l2 = list[2].split(" - ")
            if len(l2) == 2:
                game["White"] = l2[0].strip()
                game["Black"] = l2[1].strip()
            game["Result"] = list[3].strip()
        else:
            game["Result"] = "*"
        return self.rebuild_pgn(game)


# Chess-DB.com
class InternetGameChessdb(InternetGameInterface):
    def get_identity(self):
        return "Chess-DB.com", CAT_HTML

    def is_enabled(self):
        return False  # Server down

    def assign_game(self, url):
        # Verify the URL
        parsed = urlparse(url)
        if (
            parsed.netloc.lower() not in ["www.chess-db.com", "chess-db.com"]
            or "game.jsp" not in parsed.path.lower()
        ):
            return False

        # Read the arguments
        args = parse_qs(parsed.query)
        if "id" in args:
            gid = args["id"][0]
            rxp = re.compile(r"^[0-9\.]+$", re.IGNORECASE)
            if rxp.match(gid) is not None:
                self.id = gid
                return True
        return False

    def download_game(self):
        # Download
        if self.id is None:
            return None
        page = self.download("https://chess-db.com/public/game.jsp?id=%s" % self.id)
        if page is None:
            return None

        # Definition of the parser
        class chessdbparser(HTMLParser):
            def __init__(self):
                HTMLParser.__init__(self)
                self.tag_ok = False
                self.pgn = None
                self.pgn_tmp = None

            def handle_starttag(self, tag, attrs):
                if tag.lower() == "input":
                    for k, v in attrs:
                        k = k.lower()
                        if k == "name" and v == "pgn":
                            self.tag_ok = True
                        if k == "value" and v.count("[") == v.count("]"):
                            self.pgn_tmp = v

            def handle_data(self, data):
                if self.pgn is None and self.tag_ok:
                    self.pgn = self.pgn_tmp

        # Read the PGN
        parser = chessdbparser()
        parser.feed(page)
        return parser.pgn


# ChessPro.ru
class InternetGameChesspro(InternetGameInterface):
    def get_identity(self):
        return "ChessPro.ru", CAT_HTML

    def assign_game(self, url):
        return self.reacts_to(url, "chesspro.ru")

    def download_game(self):
        # Check
        if self.id is None:
            return None

        # Download the page
        page = self.download(self.id)
        if page is None:
            return None

        # Find the chess widget
        rxp = re.compile(
            r'.*OpenGame\(\s*"g[0-9]+\"\s*,"(.*)"\s*\)\s*;.*', re.IGNORECASE
        )
        lines = page.split("\n")
        for line in lines:
            m = rxp.match(line)
            if m is not None:
                return '[Annotator "ChessPro.ru"]\n%s' % m.group(1)
        return None


# Ficgs.com
class InternetGameFicgs(InternetGameInterface):
    def get_identity(self):
        return "Ficgs.com", CAT_DL

    def assign_game(self, url):
        rxp = re.compile(
            r"^https?:\/\/(\S+\.)?ficgs\.com\/game_(\d+).html", re.IGNORECASE
        )
        m = rxp.match(url)
        if m is not None:
            gid = str(m.group(2))
            if gid.isdigit() and gid != "0":
                self.id = gid
                return True
        return False

    def download_game(self):
        # Check
        if self.id is None:
            return None

        # Download
        return self.download("http://www.ficgs.com/game_%s.pgn" % self.id)


# Chessbase
class InternetGameChessbase(InternetGameInterface):
    def get_identity(self):
        return "ChessBase.com", CAT_HTML

    def assign_game(self, url):
        return self.reacts_to(url, "*")  # Any website can embed a widget from Chessbase

    def download_game(self):
        # Download
        if self.id is None:
            return None
        page = self.download(self.id)
        if page is None:
            return None

        # Definition of the parser
        class chessbaseparser(HTMLParser):
            def __init__(self):
                HTMLParser.__init__(self)
                self.tag_ok = False
                self.links = []
                self.pgn = None

            def handle_starttag(self, tag, attrs):
                # Verify the parent tag
                self.tag_ok = False
                if tag.lower() in ["p", "div"]:
                    for k, v in attrs:
                        if k.lower() == "class" and v == "cbreplay":
                            self.tag_ok = True

                    # Content by link
                    if self.tag_ok:
                        for k, v in attrs:
                            if k.lower() == "data-url":
                                self.links.append(v)

            def handle_data(self, data):
                # Content by data
                if self.tag_ok and self.pgn is None:
                    data = data.strip()
                    if len(data) > 0:  # Empty content if data-url used
                        self.pgn = data

        # Parse the page
        parser = chessbaseparser()
        parser.feed(page)
        if parser.pgn is not None:
            return parser.pgn
        else:
            parser.links = self.expand_links(parser.links, self.id)
            return self.download_list(parser.links)


# PlayOK.com
class InternetGamePlayok(InternetGameInterface):
    def get_identity(self):
        return "PlayOK.com", CAT_DL

    def assign_game(self, url):
        # Verify the hostname
        parsed = urlparse(url)
        if parsed.netloc.lower() not in ["www.playok.com", "playok.com"]:
            return False

        # Read the arguments
        args = parse_qs(parsed.query)
        if "g" in args:
            gid = args["g"][0]
            if gid[:2] == "ch":
                gid = gid[2:].replace(".txt", "")
                if gid.isdigit() and gid != "0":
                    self.id = gid
                    return True
        return False

    def download_game(self):
        if self.id is not None:
            pgn = self.download("https://www.playok.com/p/?g=ch%s.txt" % self.id)
            if len(pgn) > 16:
                return pgn
        return None


# Pychess.org
class InternetGamePychess(InternetGameInterface):
    def get_identity(self):
        return "Pychess.org", CAT_WS

    def assign_game(self, url):
        # Retrieve the ID of the game
        rxp = re.compile(
            r"https?:\/\/(www\.)?pychess(-variants\.herokuapp\.com|\.org)\/([a-z0-9]+)[\/\?\#]?",
            re.IGNORECASE,
        )
        m = rxp.match(url)
        if m is not None:
            gid = m.group(3)
            if len(gid) == 8:
                self.id = gid
                return True

        # Nothing found
        return False

    async def download_game(self):
        # Check
        if self.id is None:
            return

        # Open a websocket to retrieve the game
        async def coro(self):
            result = None
            ws = await websockets.connect(
                "wss://www.pychess.org/wsr",
                origin="https://www.pychess.org",
                ping_interval=None,
            )
            try:
                await ws.send('{"type":"board","gameId":"%s"}' % self.id)
                for i in range(5):
                    data = await ws.recv()
                    data = self.json_loads(data)
                    if data["type"] == "board" and data["gameId"] == self.id:
                        result = data["pgn"] if data["pgn"] != "" else None
                        break
            finally:
                await ws.close()
            self.data = result

        await coro(self)


# Generic
class InternetGameGeneric(InternetGameInterface):
    def get_identity(self):
        return "Generic", CAT_MISC

    def assign_game(self, url):
        # Any page is valid
        self.id = url
        return True

    def download_game(self):
        # Check
        if self.id is None:
            return None

        # Download
        req = Request(self.id, headers={"User-Agent": self.userAgent})
        response = urlopen(req)
        mime = response.info().get_content_type().lower()
        data = self.read_data(response)
        if data is None:
            return None

        # Chess file
        if (mime in ["application/x-chess-pgn", "application/pgn"]) or (
            self.allow_octet_stream and mime == "application/octet-stream"
        ):
            return data

        # Web-page
        if mime == "text/html":
            # Definition of the parser
            class linksParser(HTMLParser):
                def __init__(self):
                    HTMLParser.__init__(self)
                    self.links = []

                def handle_starttag(self, tag, attrs):
                    if tag.lower() == "a":
                        for k, v in attrs:
                            if k.lower() == "href":
                                v = v.strip()
                                u = urlparse(v)
                                if u.path.lower().endswith(".pgn"):
                                    self.links.append(v)

            # Read the links
            parser = linksParser()
            parser.feed(data)
            parser.links = self.expand_links(parser.links, self.id)
            return self.download_list(parser.links)
        return None


# Interface
chess_providers = [
    InternetGameLichess(),
    InternetGameChessgames(),
    InternetGameFicsgames(),
    InternetGameChesstempo(),
    InternetGame365chess(),
    InternetGameChesspastebin(),
    InternetGameThechessworld(),
    InternetGameChessOrg(),
    InternetGameEuropeechecs(),
    InternetGameGameknot(),
    InternetGameChessCom(),
    InternetGameSchachspielen(),
    InternetGameRedhotpawn(),
    InternetGameChesssamara(),
    InternetGame2700chess(),
    InternetGameIccf(),
    InternetGameSchacharena(),
    InternetGameChesspuzzle(),
    InternetGameChessking(),
    InternetGameIdeachess(),
    InternetGameChessdb(),
    InternetGameChesspro(),
    InternetGameFicgs(),
    InternetGameChessbase(),
    InternetGamePlayok(),
    InternetGamePychess(),
    # TODO ChessDuo.com
    InternetGameGeneric(),
]


# Get the list of chess providers
def get_internet_game_providers():
    list = [cp.get_description() for cp in chess_providers if cp.is_enabled()]
    list.sort()
    return list


# Retrieve a game from a URL
async def get_internet_game_as_pgn(url):
    # Recognize the most popular identifiers
    if url in [None, ""]:
        return None
    if re.compile(r"^[a-z0-9-]{8}$", re.IGNORECASE).match(url) is not None:
        url = "https://lichess.org/%s" % url
    elif url.isdigit():
        url = "https://www.chess.com/live/game/%s" % url

    # Check the format
    p = urlparse(url.strip())
    if "" in [p.scheme, p.netloc]:
        return None
    log.debug("URL to retrieve: %s" % url)

    # Call the chess providers
    for prov in chess_providers:
        if not prov.is_enabled():
            continue
        prov.reset()
        if prov.assign_game(url):
            # Download
            log.debug("Responding chess provider: %s" % prov.get_description())
            try:
                if prov.is_async():
                    await prov.download_game()
                    pgn = prov.data
                else:
                    pgn = prov.download_game()
                pgn = prov.sanitize(pgn)
            except Exception as e:
                pgn = None
                log.debug(str(e))

            # Check
            if pgn is None:
                log.debug("Download failed")
            else:
                log.debug("Successful download")
                return pgn
    return None


async def get_internet_game(url):
    if url in [None, ""]:
        return False
    else:
        data = await get_internet_game_as_pgn(url.strip())
        return newGameDialog.loadPgnAndRun(data)


# print(get_internet_game_as_pgn(''))



import asyncio
import os

from pychess.System.prefix import addDataPrefix
from pychess.Utils.const import WHITE, BLACK, LOCAL, WAITING_TO_START, HINT, LESSON
from pychess.Utils.LearnModel import LearnModel
from pychess.Utils.TimeModel import TimeModel
from pychess.Players.Human import Human
from pychess.System import conf
from pychess.perspectives import perspective_manager
from pychess.perspectives.learn.generateLessonsSidepanel import generateLessonsSidepanel
from pychess.perspectives.learn import lessons_solving_progress
from pychess.perspectives.learn.PuzzlesPanel import start_puzzle_game
from pychess.Savers.pgn import PGNFile
from pychess.System.protoopen import protoopen
from pychess.Players.engineNest import discoverer

__title__ = _("Lessons")

__icon__ = addDataPrefix("glade/panel_book.svg")

__desc__ = _('Guided interactive lessons in "guess the move" style')


LESSONS = []
for elem in sorted(os.listdir(path=addDataPrefix("learn/lessons/"))):
    if elem.startswith("lichess_study") and elem.endswith(".pgn"):
        title = elem.replace("beta-lichess-practice-", "")
        title = title[14 : title.find("_by_")].replace("-", " ").capitalize()
        LESSONS.append((elem, title, "lichess.org"))
    elif elem.endswith(".pgn"):
        LESSONS.append((elem, elem.replace("-", " ").capitalize(), "pychess.org"))

# Note: Find the declaration of the class Sidepanel at the end of the file


def start_lesson_from(filename, index=None):
    chessfile = PGNFile(protoopen(addDataPrefix("learn/lessons/%s" % filename)))
    chessfile.limit = 1000
    chessfile.init_tag_database()
    records, plys = chessfile.get_records()

    progress = lessons_solving_progress.get(filename, [0] * chessfile.count)

    if index is None:
        try:
            index = progress.index(0)
        except ValueError:
            index = 0

    rec = records[index]

    timemodel = TimeModel(0, 0)
    gamemodel = LearnModel(timemodel)

    chessfile.loadToModel(rec, -1, gamemodel)

    if len(gamemodel.moves) > 0:
        start_lesson_game(gamemodel, filename, chessfile, records, index, rec)
    else:
        start_puzzle_game(gamemodel, filename, records, index, rec, from_lesson=True)


def start_lesson_game(gamemodel, filename, chessfile, records, index, rec):
    gamemodel.set_learn_data(LESSON, filename, index, len(records))

    # Lichess doesn't export some study data to .pgn like
    # Orientation, Analysis mode, Chapter pinned comment, move hint comments, general fail comment
    if filename.startswith(
        "lichess_study_beta-lichess-practice-checkmating-with-a-knight-and-bishop"
    ):
        if index in (4, 6, 8, 9):
            gamemodel.tags["Orientation"] = "White"
            print(index, '[Orientation "White"]')

    color = gamemodel.boards[0].color
    player_name = conf.get("firstName")

    w_name = player_name if color == WHITE else "PyChess"
    b_name = "PyChess" if color == WHITE else player_name

    p0 = (LOCAL, Human, (WHITE, w_name), w_name)
    p1 = (LOCAL, Human, (BLACK, b_name), b_name)

    def learn_success(gamemodel):
        gamemodel.scores = {}
        chessfile.loadToModel(rec, -1, gamemodel)
        progress = lessons_solving_progress[gamemodel.source]
        progress[gamemodel.current_index] = 1
        lessons_solving_progress[gamemodel.source] = progress
        if "FEN" in gamemodel.tags:
            asyncio.create_task(gamemodel.restart_analyzer(HINT))

    gamemodel.connect("learn_success", learn_success)

    def on_game_started(gamemodel):
        perspective.activate_panel("annotationPanel")
        if "FEN" in gamemodel.tags:
            asyncio.create_task(
                gamemodel.start_analyzer(HINT, force_engine=discoverer.getEngineLearn())
            )

    gamemodel.connect("game_started", on_game_started)

    gamemodel.status = WAITING_TO_START
    perspective = perspective_manager.get_perspective("games")
    asyncio.create_task(perspective.generalStart(gamemodel, p0, p1))


# Sidepanel is a class
Sidepanel = generateLessonsSidepanel(
    solving_progress=lessons_solving_progress,
    learn_category_id=LESSON,
    entries=LESSONS,
    start_from=start_lesson_from,
)



from gi.repository import Gtk, Gdk

from pychess.System import conf
from pychess.System.prefix import addDataPrefix
from pychess.Utils.const import BLACK
from pychess.Utils.Move import toSAN, toFAN
from pychess.widgets.Background import hexcol

__title__ = _("Move History")
__active__ = True
__icon__ = addDataPrefix("glade/panel_moves.svg")
__desc__ = _(
    "The moves sheet keeps track of the players' moves and lets you navigate through the game history"
)


class Sidepanel:
    def load(self, gmwidg):
        self.gamemodel = gmwidg.board.view.model
        self.model_cids = [
            self.gamemodel.connect_after("game_changed", self.game_changed),
            self.gamemodel.connect_after("game_started", self.game_started),
            self.gamemodel.connect_after("moves_undone", self.moves_undone),
            self.gamemodel.connect_after("game_terminated", self.on_game_terminated),
        ]

        self.tv = Gtk.TreeView()
        self.tv.set_headers_visible(False)
        self.tv.set_grid_lines(True)
        self.tv.set_activate_on_single_click(True)
        self.tv.get_selection().set_mode(Gtk.SelectionMode.NONE)
        self.activated_cell = (None, None)

        def is_row_separator(treemodel, treeiter):
            mvcount, wmove, bmove, row, wbg, bbg = self.store[treeiter]
            return row == 0

        self.tv.set_row_separator_func(is_row_separator)

        self.tv.connect("style-updated", self.on_style_updated)
        movetext_font = conf.get("movetextFont")

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("mvcount", renderer, text=0)
        column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        self.tv.append_column(column)

        self.white_renderer = Gtk.CellRendererText()
        self.white_renderer.set_property("xalign", 1)
        self.white_renderer.set_property("font", movetext_font)
        self.white_column = Gtk.TreeViewColumn(
            "white", self.white_renderer, text=1, background=4
        )
        self.white_column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        self.tv.append_column(self.white_column)

        self.black_renderer = Gtk.CellRendererText()
        self.black_renderer.set_property("xalign", 1)
        self.black_renderer.set_property("font", movetext_font)
        self.black_column = Gtk.TreeViewColumn(
            "black", self.black_renderer, text=2, background=5
        )
        self.black_column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        self.tv.append_column(self.black_column)

        # To prevent black moves column expand to the right we add a dummy column finally
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("dummy", renderer)
        self.tv.append_column(column)

        scrollwin = Gtk.ScrolledWindow()
        scrollwin.add(self.tv)

        # Our liststore elements will be:
        # mvcount, white move, black move, row number, white move background, black move background
        self.store = Gtk.ListStore(str, str, str, int, str, str)
        self.tv.set_model(self.store)
        self.tv_cid = self.tv.connect("row_activated", self.on_row_activated)

        self.boardview = gmwidg.board.view
        self.cid = self.boardview.connect("shownChanged", self.shownChanged)
        scrollwin.show_all()

        self.figuresInNotation = conf.get("figuresInNotation")

        def figuresInNotationCallback(none):
            game = self.boardview.model
            if game.lesson_game:
                return

            self.figuresInNotation = conf.get("figuresInNotation")

            for i, move in enumerate(game.moves):
                board = game.variations[0][i]
                ply = game.lowply + i + 1
                if conf.get("figuresInNotation"):
                    notat = toFAN(board, move)
                else:
                    notat = toSAN(board, move, True)

                row, column = self.ply_to_row_col(ply)

                col = 2 if column == self.black_column else 1
                treeiter = self.store.get_iter(Gtk.TreePath(row))
                self.store.set_value(treeiter, col, notat)

        def font_changed(none):
            movetext_font = conf.get("movetextFont")
            self.black_renderer.set_property("font", movetext_font)
            self.white_renderer.set_property("font", movetext_font)
            self.shownChanged(self.boardview, self.boardview.shown)

        self.cids_conf = []
        self.cids_conf.append(conf.notify_add("movetextFont", font_changed))
        self.cids_conf.append(
            conf.notify_add("figuresInNotation", figuresInNotationCallback)
        )

        return scrollwin

    def get_background_rgba(self, selected=False):
        if selected:
            found, color = self.tv.get_style_context().lookup_color(
                "theme_selected_bg_color"
            )
        else:
            found, color = self.tv.get_style_context().lookup_color("theme_base_color")
        return hexcol(Gdk.RGBA(color.red, color.green, color.blue, 1))

    def on_style_updated(self, widget):
        for row in self.store:
            row[4] = self.get_background_rgba()
            row[5] = self.get_background_rgba()
        # update selected cell
        self.shownChanged(self.boardview, self.boardview.shown)

    def on_game_terminated(self, model):
        self.tv.disconnect(self.tv_cid)
        for cid in self.model_cids:
            self.gamemodel.disconnect(cid)
        self.boardview.disconnect(self.cid)
        for cid in self.cids_conf:
            conf.notify_remove(cid)

    def on_row_activated(self, tv, path, col, from_show_changed=False):
        if col not in (self.white_column, self.black_column):
            return

        # Make previous activated cell background color unselected
        old_row, old_col = self.activated_cell
        if old_row is not None:
            bg_col = 5 if old_col == self.black_column else 4
            treeiter = self.store.get_iter(Gtk.TreePath(old_row))
            self.store.set_value(
                treeiter, bg_col, self.get_background_rgba(selected=False)
            )

        # Make activated cell background color selected
        self.activated_cell = path[0], col
        bg_col = 5 if col == self.black_column else 4
        treeiter = self.store.get_iter(Gtk.TreePath(path[0]))
        self.store.set_value(treeiter, bg_col, self.get_background_rgba(selected=True))

        index = path[0] * 2 - 1 + (1 if col == self.black_column else 0)
        if self.gamemodel.starting_color == BLACK:
            index -= 1

        if index < len(self.gamemodel.boards):
            # Don't set shown board if on_row_activated() was called from shownChanged()
            if not from_show_changed:
                board = self.gamemodel.boards[index]
                self.boardview.setShownBoard(board)

    def shownChanged(self, boardview, shown):
        if boardview is None or self.gamemodel is None:
            return
        if not boardview.shownIsMainLine():
            return

        row, column = self.ply_to_row_col(shown)

        try:
            self.on_row_activated(
                self, Gtk.TreePath(row), column, from_show_changed=True
            )
            self.tv.scroll_to_cell(row)
        except ValueError:
            pass
            # deleted variations by moves_undoing

    def moves_undone(self, gamemodel, moves):
        for i in range(moves):
            treeiter = self.store.get_iter((len(self.store) - 1,))
            # If latest move is black move don't remove whole line!
            if self.store[-1][2]:
                self.store.set_value(treeiter, 2, "")
            else:
                self.store.remove(treeiter)

    def game_changed(self, gamemodel, ply):
        if self.boardview is None or self.boardview.model is None:
            return

        if len(self.store) == 0:
            for i in range(len(self.store) + gamemodel.lowply, ply + 1):
                self.add_move(gamemodel, i)
        else:
            self.add_move(gamemodel, ply)

        self.shownChanged(self.boardview, ply)

    def game_started(self, game):
        if game.lesson_game:
            return
        self.game_changed(game, game.ply)

    def add_move(self, gamemodel, ply):
        if ply == gamemodel.lowply:
            self.store.append(
                [
                    "%4s." % gamemodel.lowply,
                    "1234567",
                    "1234567",
                    0,
                    self.get_background_rgba(),
                    self.get_background_rgba(),
                ]
            )
            return

        if self.figuresInNotation:
            notat = toFAN(
                gamemodel.getBoardAtPly(ply - 1), gamemodel.getMoveAtPly(ply - 1)
            )
        else:
            notat = toSAN(
                gamemodel.getBoardAtPly(ply - 1),
                gamemodel.getMoveAtPly(ply - 1),
                localRepr=True,
            )

        row, column = self.ply_to_row_col(ply)

        if len(self.store) - 1 < row:
            mvcount = "%s." % ((ply + 1) // 2)
            if column == self.white_column:
                self.store.append(
                    [
                        mvcount,
                        notat,
                        "",
                        row,
                        self.get_background_rgba(),
                        self.get_background_rgba(),
                    ]
                )
            else:
                self.store.append(
                    [
                        mvcount,
                        "",
                        notat,
                        row,
                        self.get_background_rgba(),
                        self.get_background_rgba(),
                    ]
                )
        else:
            treeiter = self.store.get_iter(Gtk.TreePath(row))
            col = 1 if column == self.white_column else 2
            self.store.set_value(treeiter, col, notat)

    def ply_to_row_col(self, ply):
        col = ply & 1 and self.white_column or self.black_column
        if self.gamemodel.lowply & 1:
            row = (ply - self.gamemodel.lowply) // 2
        else:
            row = (ply - self.gamemodel.lowply - 1) // 2
        return row + 1, col



import asyncio
import collections
import datetime
import os
import random
import traceback
from queue import Queue
from io import StringIO

from gi.repository import GObject

from pychess.Savers.ChessFile import LoadingError
from pychess.Players.Player import (
    PlayerIsDead,
    PassInterrupt,
    TurnInterrupt,
    InvalidMove,
    GameEnded,
)
from pychess.System import conf
from pychess.System.protoopen import protoopen, protosave
from pychess.System.Log import log
from pychess.Utils.book import getOpenings
from pychess.Utils.Move import Move
from pychess.Utils.eco import get_eco
from pychess.Utils.Offer import Offer
from pychess.Utils.TimeModel import TimeModel
from pychess.Utils.DecisionSupportAlgorithm import DecisionSupportAlgorithm
from pychess.Savers import html, txt
from pychess.Variants.normal import NormalBoard

from .logic import getStatus, isClaimableDraw, playerHasMatingMaterial
from .const import (
    WAITING_TO_START,
    UNKNOWN_REASON,
    WHITE,
    ARTIFICIAL,
    RUNNING,
    FLAG_CALL,
    BLACK,
    KILLED,
    ANALYZING,
    LOCAL,
    REMOTE,
    PAUSED,
    HURRY_ACTION,
    CHAT_ACTION,
    RESIGNATION,
    BLACKWON,
    WHITEWON,
    DRAW_CALLFLAG,
    WON_RESIGN,
    DRAW,
    WON_CALLFLAG,
    DRAW_WHITEINSUFFICIENTANDBLACKTIME,
    DRAW_OFFER,
    TAKEBACK_OFFER,
    DRAW_BLACKINSUFFICIENTANDWHITETIME,
    ACTION_ERROR_NOT_OUT_OF_TIME,
    OFFERS,
    ACTION_ERROR_TOO_LARGE_UNDO,
    ACTION_ERROR_NONE_TO_WITHDRAW,
    DRAW_AGREE,
    ACTION_ERROR_NONE_TO_DECLINE,
    ADJOURN_OFFER,
    ADJOURNED,
    ABORT_OFFER,
    ABORTED,
    ADJOURNED_AGREEMENT,
    PAUSE_OFFER,
    RESUME_OFFER,
    ACTION_ERROR_NONE_TO_ACCEPT,
    ABORTED_AGREEMENT,
    WHITE_ENGINE_DIED,
    BLACK_ENGINE_DIED,
    WON_ADJUDICATION,
    UNDOABLE_STATES,
    DRAW_REPETITION,
    UNDOABLE_REASONS,
    UNFINISHED_STATES,
    DRAW_50MOVES,
    HINT,
)


class GameModel(GObject.GObject):
    """GameModel contains all available data on a chessgame.
    It also has the task of controlling players actions and moves"""

    __gsignals__ = {
        # game_started is emitted when control is given to the players for the
        # first time. Notice this is after players.start has been called.
        "game_started": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # game_changed is emitted when a move has been made.
        "game_changed": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        # moves_undoig is emitted when a undoMoves call has been accepted, but
        # before any work has been done to execute it.
        "moves_undoing": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        # moves_undone is emitted after n moves have been undone in the
        # gamemodel and the players.
        "moves_undone": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        # variation_undoig is emitted when a undo_in_variation call has been started, but
        # before any work has been done to execute it.
        "variation_undoing": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # variation_undone is emitted after 1 move have been undone in the
        # boardview shown variation
        "variation_undone": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # game_unended is emitted if moves have been undone, such that the game
        # which had previously ended, is now again active.
        "game_unended": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # game_loading is emitted if the GameModel is about to load in a chess
        # game from a file.
        "game_loading": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        # game_loaded is emitted after the chessformat handler has loaded in
        # all the moves from a file to the game model.
        "game_loaded": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        # game_saved is emitted in the end of model.save()
        "game_saved": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        # game_ended is emitted if the models state has been changed to an
        # "ended state"
        "game_ended": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        # game_terminated is emitted if the game was terminated. That is all
        # players and clocks were stopped, and it is no longer possible to
        # resume the game, even by undo.
        "game_terminated": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # game_paused is emitted if the game was successfully paused.
        "game_paused": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # game_paused is emitted if the game was successfully resumed from a
        # pause.
        "game_resumed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # action_error is currently only emitted by ICGameModel, in the case
        # the "web model" didn't accept the action you were trying to do.
        "action_error": (GObject.SignalFlags.RUN_FIRST, None, (object, int)),
        # players_changed is emitted if the players list was changed.
        "players_changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "analyzer_added": (GObject.SignalFlags.RUN_FIRST, None, (object, str)),
        "analyzer_removed": (GObject.SignalFlags.RUN_FIRST, None, (object, str)),
        "analyzer_paused": (GObject.SignalFlags.RUN_FIRST, None, (object, str)),
        "analyzer_resumed": (GObject.SignalFlags.RUN_FIRST, None, (object, str)),
        # opening_changed is emitted if the move changed the opening.
        "opening_changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # variation_added is emitted if a variation was added.
        "variation_added": (GObject.SignalFlags.RUN_FIRST, None, (object, object)),
        # variation_extended is emitted if a new move was added to a variation.
        "variation_extended": (GObject.SignalFlags.RUN_FIRST, None, (object, object)),
        # scores_changed is emitted if the analyzing scores was changed.
        "analysis_changed": (GObject.SignalFlags.RUN_FIRST, None, (int,)),
        # analysis_finished is emitted if the game analyzing finished stepping on all moves.
        "analysis_finished": (GObject.SignalFlags.RUN_FIRST, None, ()),
        # FICS games can get kibitz/whisper messages
        "message_received": (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
        # FICS games can have observers
        "observers_received": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
    }

    def __init__(self, timemodel=None, variant=NormalBoard):
        GObject.GObject.__init__(self)
        self.daemon = True
        self.variant = variant
        self.boards = [variant(setup=True)]

        self.moves = []
        self.scores = {}
        self.spy_scores = {}
        self.players = []

        self.gameno = None
        self.variations = [self.boards]

        self.terminated = False
        self.status = WAITING_TO_START
        self.reason = UNKNOWN_REASON
        self.curColor = WHITE

        # support algorithm for new players
        # type apparent : DecisionSupportAlgorithm
        self.support_algorithm = DecisionSupportAlgorithm()

        if timemodel is None:
            self.timemodel = TimeModel()
        else:
            self.timemodel = timemodel
        self.timemodel.gamemodel = self

        self.connections = collections.defaultdict(list)  # mainly for IC subclasses
        self.analyzer_cids = {}
        self.examined = False

        now = datetime.datetime.now()
        self.tags = collections.defaultdict(str)
        self.tags["Event"] = _("Local Event")
        self.tags["Site"] = _("Local Site")
        self.tags["Date"] = "%04d.%02d.%02d" % (now.year, now.month, now.day)
        self.tags["Round"] = "1"

        self.endstatus = None
        self.zero_reached_cid = None

        self.timed = self.timemodel.minutes != 0 or self.timemodel.gain != 0
        if self.timed:
            self.zero_reached_cid = self.timemodel.connect(
                "zero_reached", self.zero_reached
            )
            if self.timemodel.moves == 0:
                self.tags["TimeControl"] = "%d%s%d" % (
                    self.timemodel.minutes * 60,
                    "+" if self.timemodel.gain >= 0 else "-",
                    abs(self.timemodel.gain),
                )
            else:
                self.tags["TimeControl"] = "%d/%d" % (
                    self.timemodel.moves,
                    self.timemodel.minutes * 60,
                )
            # Notice: tags["WhiteClock"] and tags["BlackClock"] are never set
            # on the gamemodel, but simply written or read during saving/
            # loading from pgn. If you want to know the time left for a player,
            # check the time model.

        # Keeps track of offers, so that accepts can be spotted
        self.offers = {}

        # True if the game has been changed since last save
        self.needsSave = False

        # The uri the current game was loaded from, or None if not a loaded game
        self.uri = None

        # Link to additiona info
        self.info = None

        self.spectators = {}

        self.undoQueue = Queue()

        # learn_type set by LearnModel.set_learn_data()
        self.offline_lecture = False
        self.puzzle_game = False
        self.lesson_game = False
        self.end_game = False
        self.solved = False

    @property
    def practice_game(self):
        return self.puzzle_game or self.end_game

    @property
    def starting_color(self):
        return (
            BLACK
            if "FEN" in self.tags
            and self.tags["FEN"] != ""
            and self.tags["FEN"].split()[1] == "b"
            else WHITE
        )

    @property
    def orientation(self):
        if "Orientation" in self.tags:
            return BLACK if self.tags["Orintation"].lower() == "black" else WHITE
        else:
            return self.starting_color

    def zero_reached(self, timemodel, color):
        if conf.get("autoCallFlag"):
            if self.status == RUNNING and timemodel.getPlayerTime(color) <= 0:
                log.info(
                    "Automatically sending flag call on behalf of player %s."
                    % self.players[1 - color].name
                )
                self.players[1 - color].emit("offer", Offer(FLAG_CALL))

    def __repr__(self):
        string = "<GameModel at %s" % id(self)
        string += " (ply=%s" % self.ply
        if len(self.moves) > 0:
            string += ", move=%s" % self.moves[-1]
        string += ", variant=%s" % self.variant.name.encode("utf-8")
        string += f", status={str(self.status)}, reason={str(self.reason)}"
        string += ", players=%s" % str(self.players)
        string += ", tags=%s" % str(self.tags)
        if len(self.boards) > 0:
            string += "\nboard=%s" % self.boards[-1]
        return string + ")>"

    @property
    def display_text(self):
        if self.variant == NormalBoard and not self.timed:
            return "[ " + _("Untimed") + " ]"
        else:
            text = "[ "
            if self.variant != NormalBoard:
                text += self.variant.name + " "
            if self.timed:
                text += self.timemodel.display_text + " "
            return text + "]"

    def setPlayers(self, players):
        log.debug("GameModel.setPlayers: starting")
        assert self.status == WAITING_TO_START
        self.players = players
        for player in self.players:
            self.connections[player].append(player.connect("offer", self.offerReceived))
            self.connections[player].append(
                player.connect("withdraw", self.withdrawReceived)
            )
            self.connections[player].append(
                player.connect("decline", self.declineReceived)
            )
            self.connections[player].append(
                player.connect("accept", self.acceptReceived)
            )
        self.tags["White"] = str(self.players[WHITE])
        self.tags["Black"] = str(self.players[BLACK])
        log.debug("GameModel.setPlayers: -> emit players_changed")
        self.emit("players_changed")
        log.debug("GameModel.setPlayers: <- emit players_changed")
        log.debug("GameModel.setPlayers: returning")

        # when the players are set, it is known whether or not there is a bot
        # we activate the support algorithm if there is one
        # boolean to know if the game is against a bot
        # activate support algorithm if that is the case
        if self.isLocalGame():
            self.support_algorithm.set_foe_as_bot()

    def color(self, player):
        if player is self.players[0]:
            return WHITE
        else:
            return BLACK

    async def start_analyzer(self, analyzer_type, force_engine=None):
        # Don't start regular analyzers
        if os.environ.get("PYCHESS_UNITTEST") or (
            (self.practice_game or self.lesson_game)
            and force_engine is None
            and not self.solved
        ):
            return

        # prevent starting new analyzers again and again
        # when fics lecture reuses the same gamemodel
        if analyzer_type in self.spectators:
            return

        from pychess.Players.engineNest import init_engine

        analyzer = await init_engine(analyzer_type, self, force_engine=force_engine)
        if analyzer is None:
            return

        analyzer.setOptionInitialBoard(self)
        # Enable to find alternate hint in learn perspective puzzles
        if force_engine is not None:
            analyzer.setOption("MultiPV", 3)
            analyzer.analysis_depth = 20

        self.spectators[analyzer_type] = analyzer
        self.emit("analyzer_added", analyzer, analyzer_type)
        self.analyzer_cids[analyzer_type] = analyzer.connect("analyze", self.on_analyze)

    def remove_analyzer(self, analyzer_type):
        try:
            analyzer = self.spectators[analyzer_type]
        except KeyError:
            return

        analyzer.disconnect(self.analyzer_cids[analyzer_type])
        analyzer.end(KILLED, UNKNOWN_REASON)
        self.emit("analyzer_removed", analyzer, analyzer_type)
        del self.spectators[analyzer_type]

    def resume_analyzer(self, analyzer_type):
        try:
            analyzer = self.spectators[analyzer_type]
        except KeyError:
            return

        analyzer.resume()
        self.emit("analyzer_resumed", analyzer, analyzer_type)

    def pause_analyzer(self, analyzer_type):
        try:
            analyzer = self.spectators[analyzer_type]
        except KeyError:
            return

        analyzer.pause()
        self.emit("analyzer_paused", analyzer, analyzer_type)

    async def restart_analyzer(self, analyzer_type):
        self.remove_analyzer(analyzer_type)
        await self.start_analyzer(analyzer_type)

    def on_analyze(self, analyzer, analysis):
        def safe_int(p):
            if p in [None, ""]:
                return 0
            try:
                return int(p)
            except ValueError:
                return 0

        if analysis and (self.practice_game or self.lesson_game):
            for i, anal in enumerate(analysis):
                if anal is not None:
                    ply, pv, score, depth, nps = anal
                    if len(pv) > 0:
                        if ply not in self.hints:
                            self.hints[ply] = []

                        if len(self.hints[ply]) < i + 1:
                            self.hints[ply].append((pv[0], score))
                        else:
                            # Some puzzle need manual hints added by LearnModel.add_hints()
                            # because depth 20 is not enough to find the best move
                            # we have to avoid overwrite it!
                            if score > self.hints[ply][i][1]:
                                self.hints[ply][i] = (pv[0], score)
        if analysis and analysis[0] is not None:
            ply, pv, score, depth, nps = analysis[0]
            if score is not None and depth:
                if analyzer.mode == ANALYZING:
                    if (ply not in self.scores) or (
                        safe_int(self.scores[ply][2]) <= safe_int(depth)
                    ):
                        self.scores[ply] = (pv, score, depth)
                        self.emit("analysis_changed", ply)
                else:
                    if (ply not in self.spy_scores) or (
                        safe_int(self.spy_scores[ply][2]) <= safe_int(depth)
                    ):
                        self.spy_scores[ply] = (pv, score, depth)

    def setOpening(self, ply=None, redetermine=False):
        if ply is None:
            ply = self.ply

        opening = None
        while ply >= self.lowply:
            opening = get_eco(self.getBoardAtPly(ply).board.hash, exactPosition=True)
            if opening is None and redetermine:
                ply = ply - 1
            else:
                break

        if opening is not None:
            self.tags["ECO"] = opening[0]
            self.tags["Opening"] = opening[1]
            self.tags["Variation"] = opening[2]
        else:
            if redetermine:
                if "ECO" in self.tags:
                    del self.tags["ECO"]
                if "Opening" in self.tags:
                    del self.tags["Opening"]
                if "Variation" in self.tags:
                    del self.tags["Variation"]
        self.emit("opening_changed")

    # Board stuff

    def _get_ply(self):
        return self.boards[-1].ply

    ply = property(_get_ply)

    def _get_lowest_ply(self):
        return self.boards[0].ply

    lowply = property(_get_lowest_ply)

    def _get_curplayer(self):
        try:
            return self.players[self.getBoardAtPly(self.ply).color]
        except IndexError:
            log.error(f"{self.players} {self.getBoardAtPly(self.ply).color}")
            raise

    curplayer = property(_get_curplayer)

    def _get_waitingplayer(self):
        try:
            return self.players[1 - self.getBoardAtPly(self.ply).color]
        except IndexError:
            log.error(f"{self.players} {1 - self.getBoardAtPly(self.ply).color}")
            raise

    waitingplayer = property(_get_waitingplayer)

    def _plyToIndex(self, ply):
        index = ply - self.lowply
        if index < 0:
            raise IndexError(f"{ply} < {self.lowply}\n")
        return index

    def getBoardAtPly(self, ply, variation=0):
        try:
            return self.variations[variation][self._plyToIndex(ply)]
        except IndexError:
            log.error(
                "%d\t%d\t%d\t%d\t%d"
                % (self.lowply, ply, self.ply, variation, len(self.variations))
            )
            raise

    def getMoveAtPly(self, ply, variation=0):
        try:
            return Move(
                self.variations[variation][self._plyToIndex(ply) + 1].board.lastMove
            )
        except IndexError:
            log.error(
                "%d\t%d\t%d\t%d\t%d"
                % (self.lowply, ply, self.ply, variation, len(self.variations))
            )
            raise

    def hasLocalPlayer(self):
        if self.players[0].__type__ == LOCAL or self.players[1].__type__ == LOCAL:
            return True
        else:
            return False

    def hasEnginePlayer(self):
        if (
            self.players[0].__type__ == ARTIFICIAL
            or self.players[1].__type__ == ARTIFICIAL
        ):
            return True
        else:
            return False

    def isLocalGame(self):
        if self.players[0].__type__ != REMOTE and self.players[1].__type__ != REMOTE:
            return True
        else:
            return False

    def isObservationGame(self):
        return not self.hasLocalPlayer()

    def isEngine2EngineGame(self):
        if (
            len(self.players) == 2
            and self.players[0].__type__ == ARTIFICIAL
            and self.players[1].__type__ == ARTIFICIAL
        ):
            return True
        else:
            return False

    def isPlayingICSGame(self):
        if self.players and self.status in (WAITING_TO_START, PAUSED, RUNNING):
            if (
                (
                    self.players[0].__type__ == LOCAL
                    and self.players[1].__type__ == REMOTE
                )
                or (
                    self.players[1].__type__ == LOCAL
                    and self.players[0].__type__ == REMOTE
                )
                or (
                    (self.offline_lecture or self.practice_game or self.lesson_game)
                    and not self.solved
                )
                or (
                    self.players[1].__type__ == REMOTE
                    and self.players[0].__type__ == REMOTE
                    and self.examined
                    and (
                        self.players[0].name == "puzzlebot"
                        or self.players[1].name == "puzzlebot"
                    )
                    or self.players[0].name == "endgamebot"
                    or self.players[1].name == "endgamebot"
                )
            ):
                return True
        return False

    def isLoadedGame(self):
        return self.gameno is not None

    # Offer management

    def offerReceived(self, player, offer):
        log.debug(f"GameModel.offerReceived: offerer={repr(player)} {offer}")
        if player == self.players[WHITE]:
            opPlayer = self.players[BLACK]
        elif player == self.players[BLACK]:
            opPlayer = self.players[WHITE]
        else:
            # Player comments echoed to opponent if the player started a conversation
            # with you prior to observing a game the player is in #1113
            return

        if offer.type == HURRY_ACTION:
            opPlayer.hurry()

        elif offer.type == CHAT_ACTION:
            # print("GameModel.offerreceived(player, offer)", player.name, offer.param)
            opPlayer.putMessage(offer.param)

        elif offer.type == RESIGNATION:
            if player == self.players[WHITE]:
                self.end(BLACKWON, WON_RESIGN)
            else:
                self.end(WHITEWON, WON_RESIGN)

        elif offer.type == FLAG_CALL:
            assert self.timed
            if self.timemodel.getPlayerTime(1 - player.color) <= 0:
                if self.timemodel.getPlayerTime(player.color) <= 0:
                    self.end(DRAW, DRAW_CALLFLAG)
                elif not playerHasMatingMaterial(self.boards[-1], player.color):
                    if player.color == WHITE:
                        self.end(DRAW, DRAW_WHITEINSUFFICIENTANDBLACKTIME)
                    else:
                        self.end(DRAW, DRAW_BLACKINSUFFICIENTANDWHITETIME)
                else:
                    if player == self.players[WHITE]:
                        self.end(WHITEWON, WON_CALLFLAG)
                    else:
                        self.end(BLACKWON, WON_CALLFLAG)
            else:
                player.offerError(offer, ACTION_ERROR_NOT_OUT_OF_TIME)

        elif offer.type == DRAW_OFFER and isClaimableDraw(self.boards[-1]):
            reason = getStatus(self.boards[-1])[1]
            self.end(DRAW, reason)

        elif offer.type == TAKEBACK_OFFER and offer.param < self.lowply:
            player.offerError(offer, ACTION_ERROR_TOO_LARGE_UNDO)

        elif offer.type in OFFERS:
            if offer not in self.offers:
                log.debug(
                    "GameModel.offerReceived: doing {}.offer({})".format(
                        repr(opPlayer), offer
                    )
                )
                self.offers[offer] = player
                opPlayer.offer(offer)
            # If we updated an older offer, we want to delete the old one
            keys = self.offers.keys()
            for offer_ in keys:
                if offer.type == offer_.type and offer != offer_:
                    del self.offers[offer_]

    def withdrawReceived(self, player, offer):
        log.debug(f"GameModel.withdrawReceived: withdrawer={repr(player)} {offer}")
        if player == self.players[WHITE]:
            opPlayer = self.players[BLACK]
        else:
            opPlayer = self.players[WHITE]

        if offer in self.offers and self.offers[offer] == player:
            del self.offers[offer]
            opPlayer.offerWithdrawn(offer)
        else:
            player.offerError(offer, ACTION_ERROR_NONE_TO_WITHDRAW)

    def declineReceived(self, player, offer):
        log.debug(f"GameModel.declineReceived: decliner={repr(player)} {offer}")
        if player == self.players[WHITE]:
            opPlayer = self.players[BLACK]
        else:
            opPlayer = self.players[WHITE]

        if offer in self.offers and self.offers[offer] == opPlayer:
            del self.offers[offer]
            log.debug("GameModel.declineReceived: declining %s" % offer)
            opPlayer.offerDeclined(offer)
        else:
            player.offerError(offer, ACTION_ERROR_NONE_TO_DECLINE)

    def acceptReceived(self, player, offer):
        log.debug(f"GameModel.acceptReceived: accepter={repr(player)} {offer}")
        if player == self.players[WHITE]:
            opPlayer = self.players[BLACK]
        else:
            opPlayer = self.players[WHITE]

        if offer in self.offers and self.offers[offer] == opPlayer:
            if offer.type == DRAW_OFFER:
                self.end(DRAW, DRAW_AGREE)
            elif offer.type == TAKEBACK_OFFER:
                log.debug("GameModel.acceptReceived: undoMoves(%s)" % offer.param)
                self.undoMoves(offer.param)
            elif offer.type == ADJOURN_OFFER:
                self.end(ADJOURNED, ADJOURNED_AGREEMENT)
            elif offer.type == ABORT_OFFER:
                self.end(ABORTED, ABORTED_AGREEMENT)
            elif offer.type == PAUSE_OFFER:
                self.pause()
            elif offer.type == RESUME_OFFER:
                self.resume()
            del self.offers[offer]
        else:
            player.offerError(offer, ACTION_ERROR_NONE_TO_ACCEPT)

    # Data stuff

    def loadAndStart(self, uri, loader, gameno, position, first_time=True):
        if first_time:
            assert self.status == WAITING_TO_START

        uriIsFile = not isinstance(uri, str)
        if not uriIsFile:
            chessfile = loader.load(protoopen(uri))
        else:
            chessfile = loader.load(uri)

        self.gameno = gameno
        self.emit("game_loading", uri)
        try:
            chessfile.loadToModel(gameno, -1, self)
        # Postpone error raising to make games loadable to the point of the
        # error
        except LoadingError as e:
            error = e
        else:
            error = None
        if self.players:
            self.players[WHITE].setName(self.tags["White"])
            self.players[BLACK].setName(self.tags["Black"])
        self.emit("game_loaded", uri)

        self.needsSave = False
        if not uriIsFile:
            self.uri = uri
        else:
            self.uri = None

        # Even if the game "starts ended", the players should still be moved
        # to the last position, so analysis is correct, and a possible "undo"
        # will work as expected.
        for spectator in self.spectators.values():
            spectator.setOptionInitialBoard(self)
        for player in self.players:
            player.setOptionInitialBoard(self)
        if self.timed:
            self.timemodel.setMovingColor(self.boards[-1].color)

        if first_time:
            if self.status == RUNNING:
                if self.timed:
                    self.timemodel.start()

            # Store end status from Result tag
            if self.status in (DRAW, WHITEWON, BLACKWON):
                self.endstatus = self.status
            self.status = WAITING_TO_START
            self.start()

        if error:
            raise error

    def save(self, uri, saver, append, position=None, flip=False):
        if saver in (html, txt):
            fileobj = open(uri, "a" if append else "w", encoding="utf-8", newline="")
            self.uri = uri
        elif isinstance(uri, str):
            fileobj = protosave(uri, append)
            self.uri = uri
        else:
            fileobj = uri
            self.uri = None
        with fileobj:
            saver.save(fileobj, self, position, flip)
        self.needsSave = False
        self.emit("game_saved", uri)

    def get_book_move(self):
        openings = getOpenings(self.boards[-1].board)
        openings.sort(key=lambda t: t[1], reverse=True)
        if not openings:
            return None

        total_weights = 0
        for move, weight, learn in openings:
            total_weights += weight

        if total_weights < 1:
            return None

        choice = random.randint(0, total_weights - 1)

        current_sum = 0
        for move, weight, learn in openings:
            current_sum += weight
            if current_sum > choice:
                return Move(move)

    # Run stuff

    def start(self):
        async def coro():
            log.debug("GameModel.run: Starting. self=%s" % self)
            # Avoid racecondition when self.start is called while we are in
            # self.end
            if self.status != WAITING_TO_START:
                return

            if not self.isLocalGame():
                self.timemodel.handle_gain = False

            self.status = RUNNING

            for player in self.players + list(self.spectators.values()):
                event = asyncio.Event()
                is_dead = set()
                player.start(event, is_dead)

                await event.wait()

                if is_dead:
                    if player in self.players[WHITE]:
                        self.kill(WHITE_ENGINE_DIED)
                        break
                    elif player in self.players[BLACK]:
                        self.kill(BLACK_ENGINE_DIED)
                        break

            log.debug("GameModel.run: emitting 'game_started' self=%s" % self)
            self.emit("game_started")

            # Let GameModel end() itself on games started with loadAndStart()
            if not self.lesson_game:
                self.checkStatus()

            if self.isEngine2EngineGame() and self.timed:
                self.timemodel.start()
                self.timemodel.started = True

            self.curColor = self.boards[-1].color

            book_depth_max = conf.get("book_depth_max")

            while self.status in (PAUSED, RUNNING, DRAW, WHITEWON, BLACKWON):
                curPlayer = self.players[self.curColor]

                if self.timed:
                    log.debug(
                        "GameModel.run: id={}, players={}, self.ply={}: updating {}'s time".format(
                            id(self), str(self.players), str(self.ply), str(curPlayer)
                        )
                    )
                    curPlayer.updateTime(
                        self.timemodel.getPlayerTime(self.curColor),
                        self.timemodel.getPlayerTime(1 - self.curColor),
                    )
                try:
                    log.debug(
                        "GameModel.run: id={}, players={}, self.ply={}: calling {}.makeMove()".format(
                            id(self), str(self.players), self.ply, str(curPlayer)
                        )
                    )

                    move = None
                    # if the current player is a bot, get a move from the book
                    if (
                        curPlayer.__type__ == ARTIFICIAL
                        and book_depth_max > 0
                        and self.ply <= book_depth_max
                    ):
                        move = self.get_book_move()
                        log.debug(
                            "GameModel.run: id={}, players={}, self.ply={}: got move={} from book".format(
                                id(self), str(self.players), self.ply, move
                            )
                        )
                        if move is not None:
                            curPlayer.set_board(self.boards[-1].move(move))
                    # if no move exists yet (i.e., the book did not return a move or the player is human)
                    if move is None:
                        if self.ply > self.lowply:
                            move = await curPlayer.makeMove(
                                self.boards[-1], self.moves[-1], self.boards[-2]
                            )
                        else:
                            move = await curPlayer.makeMove(self.boards[-1], None, None)
                        log.debug(
                            "GameModel.run: id={}, players={}, self.ply={}: got move={} from {}".format(
                                id(self),
                                str(self.players),
                                self.ply,
                                move,
                                str(curPlayer),
                            )
                        )
                except PlayerIsDead as e:
                    if self.status in (WAITING_TO_START, PAUSED, RUNNING):
                        stringio = StringIO()
                        traceback.print_exc(file=stringio)
                        error = stringio.getvalue()
                        log.error(
                            "GameModel.run: A Player died: player=%s error=%s\n%s"
                            % (curPlayer, error, e)
                        )
                        if self.curColor == WHITE:
                            self.kill(WHITE_ENGINE_DIED)
                        else:
                            self.kill(BLACK_ENGINE_DIED)
                    break
                except InvalidMove as e:
                    stringio = StringIO()
                    traceback.print_exc(file=stringio)
                    error = stringio.getvalue()
                    log.error(
                        "GameModel.run: InvalidMove by player=%s error=%s\n%s"
                        % (curPlayer, error, e)
                    )
                    if self.curColor == WHITE:
                        self.end(BLACKWON, WON_ADJUDICATION)
                    else:
                        self.end(WHITEWON, WON_ADJUDICATION)
                    break
                except PassInterrupt:
                    log.debug(
                        "GameModel.run: id={}, players={}, self.ply={}: PassInterrupt".format(
                            id(self), str(self.players), self.ply
                        )
                    )
                    continue
                except TurnInterrupt:
                    log.debug(
                        "GameModel.run: id={}, players={}, self.ply={}: TurnInterrupt".format(
                            id(self), str(self.players), self.ply
                        )
                    )
                    self.curColor = self.boards[-1].color
                    continue
                except GameEnded:
                    log.debug("GameModel.run: got GameEnded exception")
                    break

                assert isinstance(move, Move), "%s" % repr(move)
                log.debug(
                    "GameModel.run: id={}, players={}, self.ply={}: applying move={}".format(
                        id(self), str(self.players), self.ply, str(move)
                    )
                )
                self.needsSave = True
                newBoard = self.boards[-1].move(move)
                newBoard.board.prev = self.boards[-1].board

                # newBoard.printPieces()
                # Variation on next move can exist from the hint panel...
                if self.boards[-1].board.next is not None:
                    newBoard.board.children = self.boards[-1].board.next.children

                self.boards = self.variations[0]
                self.boards[-1].board.next = newBoard.board
                self.boards.append(newBoard)
                self.moves.append(move)

                if self.timed:
                    self.timemodel.tap()

                if not self.terminated:
                    self.emit("game_changed", self.ply)

                for spectator in self.spectators.values():
                    if spectator.board == self.boards[-2]:
                        spectator.putMove(
                            self.boards[-1], self.moves[-1], self.boards[-2]
                        )

                if self.puzzle_game and len(self.moves) % 2 == 1:
                    status, reason = getStatus(self.boards[-1])
                    self.failed_playing_best = self.check_failed_playing_best(status)
                    if self.failed_playing_best:
                        # print("failed_playing_best() == True -> await asyncio.sleep(1.5) ")
                        # It may happen that analysis had no time to fill hints with best moves
                        # so we give him another chance with some additional time to think on it
                        self.spectators[HINT].setBoard(self.boards[-2])
                        # TODO: wait for an event (analyzer PV reaching 18 ply)
                        # instead of hard coded sleep time
                        await asyncio.sleep(1.5)
                        self.failed_playing_best = self.check_failed_playing_best(
                            status
                        )

                self.checkStatus()

                self.setOpening()

                self.curColor = 1 - self.curColor

            self.checkStatus()

        asyncio.create_task(coro())

    def checkStatus(self):
        """Updates self.status so it fits with what getStatus(boards[-1])
        would return. That is, if the game is e.g. check mated this will
        call mode.end(), or if moves have been undone from an otherwise
        ended position, this will call __resume and emit game_unended."""
        log.debug("GameModel.checkStatus:")

        # call flag by engine
        if self.isEngine2EngineGame() and self.status in UNDOABLE_STATES:
            return

        status, reason = getStatus(self.boards[-1])

        if self.practice_game and (
            len(self.moves) % 2 == 1 or status in UNDOABLE_STATES
        ):
            self.check_goal(status, reason)

        if self.endstatus is not None:
            self.end(self.endstatus, reason)
            return

        if status != RUNNING and self.status in (WAITING_TO_START, PAUSED, RUNNING):
            if status == DRAW and reason in (DRAW_REPETITION, DRAW_50MOVES):
                if self.isEngine2EngineGame():
                    self.end(status, reason)
                    return
            else:
                self.end(status, reason)
                return

        if (
            status != self.status
            and self.status in UNDOABLE_STATES
            and self.reason in UNDOABLE_REASONS
        ):
            self.__resume()
            self.status = status
            self.reason = UNKNOWN_REASON
            self.emit("game_unended")

    def __pause(self):
        log.debug("GameModel.__pause: %s" % self)
        if self.isEngine2EngineGame():
            for player in self.players:
                player.end(self.status, self.reason)
            if self.timed:
                self.timemodel.end()
        else:
            for player in self.players:
                player.pause()
            if self.timed:
                self.timemodel.pause()

    def pause(self):
        """Players will raise NotImplementedError if they doesn't support
        pause. Spectators will be ignored."""

        self.__pause()
        self.status = PAUSED
        self.emit("game_paused")

    def __resume(self):
        for player in self.players:
            player.resume()
        if self.timed:
            self.timemodel.resume()
        self.emit("game_resumed")

    def resume(self):
        self.status = RUNNING
        self.__resume()

    def end(self, status, reason):
        if self.status not in UNFINISHED_STATES:
            log.info(
                "GameModel.end: Can't end a game that's already ended: %s %s"
                % (status, reason)
            )
            return
        if self.status not in (WAITING_TO_START, PAUSED, RUNNING):
            self.needsSave = True

        log.debug(
            "GameModel.end: players=%s, self.ply=%s: Ending a game with status %d for reason %d"
            % (repr(self.players), str(self.ply), status, reason)
        )
        self.status = status
        self.reason = reason

        self.emit("game_ended", reason)

        self.__pause()

    def kill(self, reason):
        log.debug(
            "GameModel.kill: players=%s, self.ply=%s: Killing a game for reason %d\n%s"
            % (
                repr(self.players),
                str(self.ply),
                reason,
                "".join(traceback.format_list(traceback.extract_stack())).strip(),
            )
        )

        self.status = KILLED
        self.reason = reason

        for player in self.players:
            player.end(self.status, reason)

        for spectator in self.spectators.values():
            spectator.end(self.status, reason)

        if self.timed:
            self.timemodel.end()

        self.emit("game_ended", reason)

    def terminate(self):
        log.debug("GameModel.terminate: %s" % self)
        self.terminated = True

        if self.status != KILLED:
            for player in self.players:
                player.end(self.status, self.reason)

            analyzer_types = list(self.spectators.keys())
            for analyzer_type in analyzer_types:
                self.remove_analyzer(analyzer_type)

            if self.timed:
                log.debug("GameModel.terminate: -> timemodel.end()")
                self.timemodel.end()
                log.debug(
                    "GameModel.terminate: <- timemodel.end() %s" % repr(self.timemodel)
                )
                if self.zero_reached_cid is not None:
                    self.timemodel.disconnect(self.zero_reached_cid)

        # ICGameModel may did this if game was a FICS game
        if self.connections is not None:
            for player in self.players:
                for cid in self.connections[player]:
                    player.disconnect(cid)
        self.connections = {}

        self.timemodel.gamemodel = None
        self.players = []
        self.emit("game_terminated")

    # Other stuff

    def undoMoves(self, moves):
        """Undo and remove moves number of moves from the game history from
        the GameModel, players, and any spectators"""
        if self.ply < 1 or moves < 1:
            return
        if self.ply - moves < 0:
            # There is no way in the current threaded/asynchronous design
            # for the GUI to know that the number of moves it requests to takeback
            # will still be valid once the undo is actually processed. So, until
            # we either add some locking or get a synchronous design, we quietly
            # "fix" the takeback request rather than cause AssertionError or IndexError
            moves = 1

        log.debug(
            "GameModel.undoMoves: players={}, self.ply={}, moves={}, board={}".format(
                repr(self.players), self.ply, moves, self.boards[-1]
            )
        )
        self.emit("moves_undoing", moves)
        self.needsSave = True

        self.boards = self.variations[0]
        del self.boards[-moves:]
        del self.moves[-moves:]
        self.boards[-1].board.next = None

        for player in self.players:
            player.playerUndoMoves(moves, self)
        for spectator in self.spectators.values():
            spectator.spectatorUndoMoves(moves, self)

        log.debug("GameModel.undoMoves: undoing timemodel")
        if self.timed:
            self.timemodel.undoMoves(moves)

        self.checkStatus()
        self.setOpening(redetermine=True)

        self.emit("moves_undone", moves)

    def isChanged(self):
        if self.ply == 0:
            return False
        if self.needsSave:
            return True
        # what was this for?
        # if not self.uri or not isWriteable(self.uri):
        # return True
        return False

    def add_variation(self, board, moves, comment="", score="", emit=True):
        if board.board.next is None:
            # If we are in the latest played board, and want to add a variation
            # we have to add the latest move first
            if board.board.lastMove is None or board.board.prev is None:
                return
            moves = [Move(board.board.lastMove)] + moves
            board = board.board.prev.pieceBoard

        board0 = board
        board = board0.clone()
        board.board.prev = None

        # this prevents annotation panel node searches to find this instead of board0
        board.board.hash = -1

        if comment:
            board.board.children.append(comment)

        variation = [board]

        for move in moves:
            new = board.move(move)
            if len(variation) == 1:
                new.board.prev = board0.board
                variation[0].board.next = new.board
            else:
                new.board.prev = board.board
                board.board.next = new.board
            variation.append(new)
            board = new

        board0.board.next.children.append([vboard.board for vboard in variation])
        if score:
            variation[-1].board.children.append(score)

        head = None
        for vari in self.variations:
            if board0 in vari:
                head = vari
                break

        variation[0] = board0
        self.variations.append(head[: board0.ply - self.lowply] + variation)
        self.needsSave = True
        if emit:
            self.emit(
                "variation_added", board0.board.next.children[-1], board0.board.next
            )
        return self.variations[-1]

    def add_move2variation(self, board, move, variationIdx):
        new = board.move(move)
        new.board.prev = board.board
        board.board.next = new.board

        # Find the variation (low level lboard list) to append
        cur_board = board.board
        vari = None
        while cur_board.prev is not None:
            for child in cur_board.prev.next.children:
                if isinstance(child, list) and cur_board in child:
                    vari = child
                    break
            if vari is None:
                cur_board = cur_board.prev
            else:
                break
        vari.append(new.board)

        self.variations[variationIdx].append(new)
        self.needsSave = True
        self.emit("variation_extended", board.board, new.board)

    def remove_variation(self, board, parent):
        """board must be an lboard object of the first Board object of a variation Board(!) list"""
        # Remove the variation (list of lboards) containing board from parent's children list
        for child in parent.children:
            if isinstance(child, list) and board in child:
                parent.children.remove(child)
                break

        # Remove all variations from gamemodel's variations list which contains this board
        for vari in self.variations[1:]:
            if board.pieceBoard in vari:
                self.variations.remove(vari)

        # remove null_board if variation was added on last played move
        if not parent.fen_was_applied:
            parent.prev.next = None

        self.needsSave = True

    def undo_in_variation(self, board):
        """board must be the latest Board object of a variation board list"""
        assert board.board.next is None and len(board.board.children) == 0
        self.emit("variation_undoing")

        for vari in self.variations[1:]:
            if board in vari:
                break

        board = board.board
        parent = board.prev.next

        # If this is a one move only variation we have to remove the whole variation
        # if it's a longer one, just remove the latest move from it
        first_vari_moves = [
            child[1] for child in parent.children if not isinstance(child, str)
        ]
        if board in first_vari_moves:
            self.remove_variation(board, parent)
        else:
            board.prev.next = None
            del vari[-1]

        self.needsSave = True
        self.emit("variation_undone")



import re
from math import ceil
import time
import operator

from gi.repository import GLib, GObject

from pychess.System.Log import log
from pychess.ic import GAME_TYPES, BLKCMD_ALLOBSERVERS

titles = r"(?:\([A-Z*]+\))*"
names = "([A-Za-z]+)" + titles
titlesC = re.compile(titles)
namesC = re.compile(names)
ratings = r"\(\s*([0-9\ \-\+]{1,4}[P E]?|UNR)\)"

CHANNEL_SHOUT = "shout"
CHANNEL_CSHOUT = "cshout"


class ChatManager(GObject.GObject):
    __gsignals__ = {
        "channelMessage": (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str, bool, bool, str, str),
        ),
        "kibitzMessage": (GObject.SignalFlags.RUN_FIRST, None, (str, int, str)),
        "whisperMessage": (GObject.SignalFlags.RUN_FIRST, None, (str, int, str)),
        "privateMessage": (GObject.SignalFlags.RUN_FIRST, None, (str, str, bool, str)),
        "bughouseMessage": (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
        "announcement": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "arrivalNotification": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "departedNotification": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "channelAdd": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "channelRemove": (GObject.SignalFlags.RUN_FIRST, None, (str,)),
        "channelJoinError": (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
        "channelsListed": (GObject.SignalFlags.RUN_FIRST, None, (object,)),
        "channelLog": (GObject.SignalFlags.RUN_FIRST, None, (str, int, str, str)),
        "toldChannel": (GObject.SignalFlags.RUN_FIRST, None, (str, int)),
        "receivedChannels": (GObject.SignalFlags.RUN_FIRST, None, (str, object)),
        "receivedNames": (GObject.SignalFlags.RUN_FIRST, None, (str, object)),
        "observers_received": (GObject.SignalFlags.RUN_FIRST, None, (str, str)),
    }

    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection

        self.connection.expect_line(
            self.onPrivateMessage,
            r"%s(\*)?(?:\[\d+\])? (?:tells you|says): (.*)" % names,
        )
        self.connection.expect_line(
            self.onAnnouncement, r"\s*\*\*ANNOUNCEMENT\*\* (.*)"
        )
        self.connection.expect_line(
            self.onChannelMessage, r"%s(\*)?\((\d+)\): (.*)" % names
        )
        self.connection.expect_line(
            self.onShoutMessage, r"%s(\*)? (c-)?shouts: (.*)" % names
        )
        self.connection.expect_line(self.onShoutMessage, r"--> %s(\*)?:? (.*)" % names)
        self.connection.expect_line(
            self.onKibitzMessage, rf"{names}{ratings}\[(\d+)\] kibitzes: (.*)"
        )
        self.connection.expect_line(
            self.onWhisperMessage,
            rf"{names}{ratings}\[(\d+)\] whispers: (.*)",
        )

        self.connection.expect_line(
            self.onArrivalNotification, r"Notification: %s has arrived\." % names
        )
        self.connection.expect_line(
            self.onDepartedNotification, r"Notification: %s has departed\." % names
        )

        self.connection.expect_fromto(
            self.onChannelList,
            "channels only for their designated topics.",
            "SPECIAL NOTE",
        )

        self.connection.expect_line(
            lambda m: GLib.idle_add(self.emit, "channelAdd", m.groups()[0]),
            r"\[(\d+)\] added to your channel list.",
        )
        self.connection.expect_line(
            lambda m: GLib.idle_add(self.emit, "channelRemove", m.groups()[0]),
            r"\[(\d+)\] removed to your channel list.",
        )

        self.connection.expect_line(
            lambda m: GLib.idle_add(self.emit, "channelJoinError", *m.groups()),
            r"Only (.+?) may join channel (\d+)\.",
        )

        self.connection.expect_line(
            self.getNoChannelPlayers, r"Channel (\d+) is empty\."
        )
        self.connection.expect_fromto(
            self.getChannelPlayers,
            r"Channel (\d+)(?: \"(\w+)\")?: (.+)",
            r"(\d+) player(?: is|s are) in channel \d+\.",
        )

        self.connection.expect_fromto(
            self.gotPlayerChannels,
            "%s is in the following channels:" % names,
            r"(?!(?:\d+\s+)+)",
        )

        # self.connection.expect_line (self.toldChannel,
        #        '\(told (\d+) players in channel (\d+) ".+"\)')
        # (told Chronatog)

        # Only chess advisers and admins may join channel 63.
        # Only (.+?) may sey send tells to channel (\d+).
        # Only admins may send tells to channel 0.
        # Only registered users may send tells to channels other than 4, 7 and 53.

        self.currentLogChannel = None
        self.connection.expect_line(
            self.onChannelLogStart,
            r":Channel (\d+|shout|c-shout) log for the last \d+ minutes:$",
        )
        self.connection.expect_line(
            self.onChannelLogLine,
            r":\[(\d+):(\d+):(\d+)\] (?:(?:--> )?%s(?: shouts)?)\S* (.+)$" % names,
        )
        self.connection.expect_line(
            self.onChannelLogBreak, ':Use "tell chLog Next" to print more.$'
        )

        # TODO handling of this case is nessesary for console:
        # fics% tell 1 hi
        # You are not in channel 1, auto-adding you if possible.

        # Setting 'Lang' is a workaround for
        # http://code.google.com/p/pychess/issues/detail?id=376
        # and can be removed when conversion to FICS block mode is done
        self.connection.client.run_command("set Lang English")

        self.connection.client.run_command("set height 240")

        self.connection.client.run_command("inchannel %s" % self.connection.username)
        self.connection.client.run_command("help channel_list")
        self.channels = {}

        # Observing 112 [DrStupp vs. hajaK]: pgg (1 user)
        self.connection.expect_line(
            self.get_allob_list, r"(?:Observing|Examining)\s+(\d+).+: (.+) \("
        )

        self.connection.expect_line(
            self.on_allob_no, r"No one is observing game (\d+)."
        )

    def get_allob_list(self, match):
        """Description: Processes the returning pattern matched of the FICS allob command
                     extracts out the gameno and a list of observers before emmiting them for collection
                     by the observers view
        match: (re.reg-ex) is the complied matching pattern for processing
        """

        obs_dic = {}
        gameno = match.group(1)
        observers = match.group(2)
        oblist = observers.split()
        for player in oblist:
            if "(U)" not in player:  # deals with unregistered players
                try:
                    if "(" in player:  # deals with admin and multi titled players
                        player, rest = player.split("(", 1)
                    ficsplayer = self.connection.players.get(player)
                    obs_dic[player] = ficsplayer.getRatingByGameType(
                        GAME_TYPES["standard"]
                    )
                except KeyError:
                    obs_dic[player] = 0
                    # print("player %s is not in self.connection.players" % player)
            else:
                obs_dic[player] = 0
        obs_sorted = sorted(obs_dic.items(), key=operator.itemgetter(1), reverse=True)
        obs_str = ""
        for toople in obs_sorted:
            player, rating = toople
            if rating == 0:
                obs_str += "%s " % player  # Don't print ratings for guest accounts
            else:
                obs_str += f"{player}({rating}) "
        self.emit("observers_received", gameno, obs_str)

    get_allob_list.BLKCMD = BLKCMD_ALLOBSERVERS

    def on_allob_no(self, match):
        gameno = match.group(1)
        self.emit("observers_received", gameno, "")

    on_allob_no.BLKCMD = BLKCMD_ALLOBSERVERS

    def getChannels(self):
        return self.channels

    def getJoinedChannels(self):
        channels = self.connection.lvm.getList("channel")
        return channels

    channelListItem = re.compile(r"((?:\d+,?)+)\s*(.*)")

    def onChannelList(self, matchlist):
        self.channels = [
            (CHANNEL_SHOUT, _("Shout")),
            (CHANNEL_CSHOUT, _("Chess Shout")),
        ]
        numbers = set(range(256))  # TODO: Use limits->Server->Channels
        for line in matchlist[1:-1]:
            match = self.channelListItem.match(line)
            if not match:
                continue
            ids, desc = match.groups()
            for id in ids.split(","):
                numbers.remove(int(id))
                self.channels.append((id, desc))
        for i in numbers:
            self.channels.append((str(i), _("Unofficial channel %d" % i)))

        GLib.idle_add(self.emit, "channelsListed", self.channels)

    def getNoChannelPlayers(self, match):
        channel = match.groups()[0]
        GLib.idle_add(self.emit, "receivedNames", channel, [])

    def getChannelPlayers(self, matchlist):
        channel, name, people = matchlist[0].groups()
        people += " " + " ".join(matchlist[1:-1])
        people = namesC.findall(titlesC.sub("", people))
        GLib.idle_add(self.emit, "receivedNames", channel, people)

    def gotPlayerChannels(self, matchlist):
        list = []
        for line in matchlist[1:-1]:
            list += line.split()

    def onPrivateMessage(self, match):
        name, isadmin, text = match.groups()
        text = self.entityDecode(text)
        GLib.idle_add(self.emit, "privateMessage", name, "title", isadmin, text)

    def onAnnouncement(self, match):
        text = match.groups()[0]
        text = self.entityDecode(text)
        GLib.idle_add(self.emit, "announcement", text)

    def onChannelMessage(self, match):
        name, isadmin, channel, text = match.groups()
        text = self.entityDecode(text)
        isme = name.lower() == self.connection.username.lower()
        GLib.idle_add(self.emit, "channelMessage", name, isadmin, isme, channel, text)

    def onShoutMessage(self, match):
        if len(match.groups()) == 4:
            name, isadmin, type, text = match.groups()
        elif len(match.groups()) == 3:
            name, isadmin, text = match.groups()
            type = ""
        text = self.entityDecode(text)
        isme = name.lower() == self.connection.username.lower()
        # c-shout should be used ONLY for chess-related messages, such as
        # questions about chess or announcing being open for certain kinds of
        # chess matches. Use "shout" for non-chess messages.

        # t-shout is used to invite to tournaments
        if type == "c-":
            GLib.idle_add(
                self.emit, "channelMessage", name, isadmin, isme, CHANNEL_CSHOUT, text
            )
        else:
            GLib.idle_add(
                self.emit, "channelMessage", name, isadmin, isme, CHANNEL_SHOUT, text
            )

    def onKibitzMessage(self, match):
        name, rating, gameno, text = match.groups()
        text = self.entityDecode(text)
        GLib.idle_add(self.emit, "kibitzMessage", name, int(gameno), text)

    def onWhisperMessage(self, match):
        name, rating, gameno, text = match.groups()
        text = self.entityDecode(text)
        GLib.idle_add(self.emit, "whisperMessage", name, int(gameno), text)

    def onArrivalNotification(self, match):
        name = match.groups()[0]
        try:
            player = self.connection.players.get(name)
        except KeyError:
            return
        if player.name not in self.connection.notify_users:
            self.connection.notify_users.append(player.name)
        GLib.idle_add(self.emit, "arrivalNotification", player)

    def onDepartedNotification(self, match):
        name = match.groups()[0]
        try:
            player = self.connection.players.get(name)
        except KeyError:
            return
        GLib.idle_add(self.emit, "departedNotification", player)

    def toldChannel(self, match):
        amount, channel = match.groups()
        GLib.idle_add(self.emit, "toldChannel", channel, int(amount))

    def onChannelLogStart(self, match):
        (channel,) = match.groups()
        self.currentLogChannel = channel

    def onChannelLogLine(self, match):
        if not self.currentLogChannel:
            log.warning("Received log line before channel was set")
            return
        hour, minutes, secs, handle, text = match.groups()
        conv_time = self.convTime(int(hour), int(minutes), int(secs))
        text = self.entityDecode(text)
        GLib.idle_add(
            self.emit, "channelLog", self.currentLogChannel, conv_time, handle, text
        )

    def onChannelLogBreak(self, match):
        self.connection.client.run_command("xtell chlog Next")

    def convTime(self, h, m, s):
        # Convert to timestamp
        t1, t2, t3, t4, t5, t6, t7, t8, t9 = time.localtime()
        tstamp = time.mktime((t1, t2, t3, h, m, s, 0, 0, 0))
        # Difference to now in hours
        dif = (tstamp - time.time()) / 60.0 / 60.0
        # As we know there is maximum 30 minutes in difference, we can guess when the
        # message was sent, without knowing the sending time zone
        return tstamp - ceil(dif) * 60 * 60

    entityExpr = re.compile("&#x([a-f0-9]+);")

    def entityDecode(self, text):
        return self.entityExpr.sub(lambda m: chr(int(m.groups()[0], 16)), text)

    def entityEncode(self, text):
        buf = []
        for char in text:
            if not 32 <= ord(char) <= 127:
                char = "&#" + hex(ord(char))[1:] + ";"
            buf.append(char)
        return "".join(buf)

    def getChannelLog(self, channel, minutes=30):
        """Channel can be channel_id, shout or c-shout"""
        assert 1 <= minutes <= 120
        # Using the chlog bot
        self.connection.client.run_command(
            "xtell chlog show %s -t %d" % (channel, minutes)
        )

    def getPlayersChannels(self, player):
        self.connection.client.run_command("inchannel %s" % player)

    def getPeopleInChannel(self, channel):
        if channel in (CHANNEL_SHOUT, CHANNEL_CSHOUT):
            people = self.connection.players.get_online_playernames()
            GLib.idle_add(self.emit, "receivedNames", channel, people)
        self.connection.client.run_command("inchannel %s" % channel)

    def joinChannel(self, channel):
        self.connection.client.run_command("+channel %s" % channel)

    def removeChannel(self, channel):
        self.connection.client.run_command("-channel %s" % channel)

    def mayTellChannel(self, channel):
        if self.connection.isRegistred() or channel in ("4", "7", "53"):
            return True
        return False

    def tellPlayer(self, player, message):
        message = self.entityEncode(message)
        self.connection.client.run_command(f"tell {player} {message}")

    def tellChannel(self, channel, message):
        message = self.entityEncode(message)
        if channel == CHANNEL_SHOUT:
            self.connection.client.run_command("shout %s" % message)
        elif channel == CHANNEL_CSHOUT:
            self.connection.client.run_command("cshout %s" % message)
        else:
            self.connection.client.run_command(f"tell {channel} {message}")

    def tellAll(self, message):
        message = self.entityEncode(message)
        self.connection.client.run_command("shout %s" % message)

    def tellGame(self, gameno, message):
        message = self.entityEncode(message)
        self.connection.client.run_command(f"xkibitz {gameno} {message}")

    def tellOpponent(self, message):
        message = self.entityEncode(message)
        self.connection.client.run_command("say %s" % message)

    def tellBughousePartner(self, message):
        message = self.stripChars(message)
        self.connection.client.run_command("ptell %s" % message)

    def tellUser(self, player, message):
        for line in message.strip().split("\n"):
            self.tellPlayer(player, line)

    def whisper(self, message):
        message = self.entityEncode(message)
        self.connection.client.run_command("whisper %s" % message)
